# from .main import setup_bumps
from dataclasses import dataclass
import functools
import os
import signal
import socket
from typing import Callable, Dict, Optional
import warnings
from aiohttp import web, ClientSession
import asyncio
import socketio
from typing import Union, List
from pathlib import Path
import json
import re
import sys

import matplotlib

matplotlib.use("agg")

import mimetypes

mimetypes.add_type("text/css", ".css")
mimetypes.add_type("text/html", ".html")
mimetypes.add_type("application/json", ".json")
mimetypes.add_type("text/javascript", ".js")
mimetypes.add_type("text/javascript", ".mjs")
mimetypes.add_type("image/png", ".png")
mimetypes.add_type("image/svg+xml", ".svg")

from . import api
from .fit_thread import EVT_FIT_PROGRESS
from .state_hdf5_backed import SERIALIZERS, UNDEFINED
from .logger import logger, list_handler, console_handler
from . import persistent_settings

TRACE_MEMORY = False

# can get by name and not just by id

routes = web.RouteTableDef()
# sio = socketio.AsyncServer(cors_allowed_origins="*", serializer='msgpack')
sio = socketio.AsyncServer(cors_allowed_origins="*")
app = web.Application()
CLIENT_PATH = Path(__file__).parent.parent / "client"
APPLICATION_NAME = "bumps"


async def index(request):
    """Serve the client-side application."""
    index_path = CLIENT_PATH / "dist" / "index.html"
    if not index_path.exists():
        return web.Response(
            body=f"<h2>Client not built</h2>\
                <div>Please run <pre>python -m {APPLICATION_NAME}.webview.build_client</pre></div>",
            content_type="text/html",
            status=404,
        )
    return web.FileResponse(CLIENT_PATH / "dist" / "index.html")


app.router.add_get("/", index)

sio.attach(app)


def rest_get(fn):
    """
    Add a REST (GET) route for the function, which can also be used for
    """

    @routes.get(f"/{fn.__name__}")
    async def handler(request: web.Request):
        result = await fn(**request.query)
        return web.json_response(result)

    # pass the function to the next decorator unchanged...
    return fn


@sio.event
async def connect(sid: str, environ, data=None):
    for topic, contents in api.state.topics.items():
        message = contents[-1] if len(contents) > 0 else None
        if message is not None:
            await sio.emit(topic, message, to=sid)
    logger.info(f"Connected: session ID {sid}")


@sio.event
def disconnect(sid):
    logger.info(f"Disconnected: session ID {sid}")


@sio.event
async def set_base_path(sid: str, pathlist: List[str]):
    path = str(Path(*pathlist))
    persistent_settings.set_value("base_path", path, application=APPLICATION_NAME)


async def disconnect_all_clients():
    # disconnect all clients:
    clients = list(sio.manager.rooms.get("/", {None: {}}).get(None).keys())
    for client in clients:
        await sio.disconnect(client)
    while clients:
        clients = list(sio.manager.rooms.get("/", {None: {}}).get(None).keys())
        await asyncio.sleep(0.1)


async def _shutdown():
    await disconnect_all_clients()
    logger.info("webserver shutdown tasks complete")
    await asyncio.sleep(0.1)
    raise web.GracefulExit()


api._shutdown = _shutdown

import argparse


@dataclass
class BumpsOptions:
    """provide type hints for arguments"""

    filename: Optional[str] = None
    headless: bool = True
    external: bool = False
    port: int = 0
    hub: Optional[str] = None
    fit: Optional[str] = None
    start: bool = False
    read_store: Optional[str] = None
    write_store: Optional[str] = None
    store: Optional[str] = None
    exit: bool = False
    serializer: SERIALIZERS = "dill"
    trace: bool = False
    parallel: int = 0
    path: Optional[str] = None
    no_auto_history: bool = False
    convergence_heartbeat: bool = False
    use_persistent_path: bool = False


OPTIONS_CLASS = BumpsOptions


def get_commandline_options(arg_defaults: Optional[Dict] = None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "filename",
        nargs="?",
        help="problem file to load, .py or .json (serialized) fitproblem",
    )
    # parser.add_argument('-d', '--debug', action='store_true', help='autoload modules on change')
    parser.add_argument(
        "-x",
        "--headless",
        action="store_true",
        help="do not automatically load client in browser",
    )
    parser.add_argument(
        "--external",
        action="store_true",
        help="listen on all interfaces, including external (local connections only if not set)",
    )
    parser.add_argument("-p", "--port", default=0, type=int, help="port on which to start the server")
    parser.add_argument(
        "--hub",
        default=None,
        type=str,
        help="api address of parent hub (only used when called as subprocess)",
    )
    parser.add_argument(
        "--fit",
        default=None,
        type=str,
        choices=list(api.FITTERS_BY_ID.keys()),
        help="fitting engine to use; see manual for details",
    )
    parser.add_argument("--start", action="store_true", help="start fit when problem loaded")
    parser.add_argument(
        "--store",
        default=None,
        type=str,
        help="set read_store and write_store to same file",
    )
    parser.add_argument(
        "--read_store",
        default=None,
        type=str,
        help="read initial session state from file (overrides --store)",
    )
    parser.add_argument(
        "--write_store",
        default=None,
        type=str,
        help="output file for session state (overrides --store)",
    )
    parser.add_argument(
        "--exit",
        action="store_true",
        help="end process when fit complete (fit results lost unless write_store is specified)",
    )
    parser.add_argument(
        "--serializer",
        default=OPTIONS_CLASS.serializer,
        type=str,
        choices=["pickle", "dill", "dataclass"],
        help="strategy for serializing problem, will use value from store if it has already been defined",
    )
    parser.add_argument(
        "--trace",
        action="store_true",
        help="enable memory tracing (prints after every uncertainty update in dream)",
    )
    parser.add_argument(
        "--parallel",
        default=0,
        type=int,
        help="run fit using multiprocessing for parallelism; use --parallel=0 for all cpus",
    )
    parser.add_argument(
        "--path",
        default=None,
        type=str,
        help="set initial path for save and load dialogs",
    )
    parser.add_argument(
        "--no_auto_history",
        action="store_true",
        help="disable auto-appending problem state to history on load and at fit end",
    )
    parser.add_argument(
        "--convergence_heartbeat",
        action="store_true",
        help="enable convergence heartbeat for jupyter kernel (keeps kernel alive during fit)",
    )
    parser.add_argument(
        "--use_persistent_path",
        action="store_true",
        help="save most recently used path to disk for persistence between sessions",
    )
    # parser.add_argument('-c', '--config-file', type=str, help='path to JSON configuration to load')
    namespace = OPTIONS_CLASS()
    if arg_defaults is not None:
        logger.debug(f"arg_defaults: {arg_defaults}")
        for k, v in arg_defaults.items():
            setattr(namespace, k, v)
    args = parser.parse_args(namespace=namespace)
    return args


def wrap_with_sid(function: Callable):
    """
    throw away first parameter sid: str
    for compatibility with socket.io
    (none of the API functions use sid value)
    """

    @functools.wraps(function)
    async def with_sid(sid: str, *args, **kwargs):
        return await function(*args, **kwargs)

    return with_sid


def setup_sio_api():
    api.EMITTERS["socketio"] = sio.emit
    for name, action in api.REGISTRY.items():
        sio.on(name, handler=wrap_with_sid(action))
        rest_get(action)


def enable_convergence_kernel_heartbeat():
    from comm import create_comm

    comm = create_comm(target_name="heartbeat")

    async def send_heartbeat_on_convergence(event: str, *args, **kwargs):
        if event == "updated_convergence":
            comm.send({"status": "alive"})

    api.EMITTERS["convergence_heartbeat"] = send_heartbeat_on_convergence


def setup_app(sock: Optional[socket.socket] = None, options: OPTIONS_CLASS = OPTIONS_CLASS()):
    static_assets_path = CLIENT_PATH / "dist" / "assets"

    if static_assets_path.exists():
        app.router.add_static("/assets", static_assets_path)

    if options.use_persistent_path:
        api.state.base_path = persistent_settings.get_value(
            "base_path", str(Path().absolute()), application=APPLICATION_NAME
        )
    elif options.path is not None and Path(options.path).exists():
        api.state.base_path = options.path
    else:
        api.state.base_path = str(Path.cwd().absolute())

    if options.read_store is not None and options.store is not None:
        warnings.warn("read_store and store are both set; read_store will be used to initialize state")
    if options.write_store is not None and options.store is not None:
        warnings.warn("write_store and store are both set; write_store will be used to save state")

    read_store = options.read_store if options.read_store is not None else options.store
    write_store = options.write_store if options.write_store is not None else options.store
    if read_store is not None:
        read_store_path = Path(read_store).absolute()
        api.state.read_session_file(str(read_store_path))
        if write_store is None:
            api.state.shared.session_output_file = dict(
                pathlist=list(read_store_path.parent.parts),
                filename=read_store_path.name,
            )
    if write_store is not None:
        write_store_path = Path(write_store).absolute()
        api.state.shared.session_output_file = dict(
            pathlist=list(write_store_path.parent.parts), filename=write_store_path.name
        )
        api.state.shared.autosave_session = True

    if api.state.problem.serializer is None or api.state.problem.serializer == "":
        api.state.problem.serializer = options.serializer

    if options.no_auto_history:
        api.state.shared.autosave_history = False

    if options.trace:
        global TRACE_MEMORY
        TRACE_MEMORY = True
        api.TRACE_MEMORY = True

    # app.on_startup.append(lambda App: publish('', 'local_file_path', Path().absolute().parts))
    if options.fit is not None:
        app.on_startup.append(lambda App: api.state.shared.set("selected_fitter", options.fit))

    fitter_id = options.fit
    if fitter_id is None:
        fitter_id = api.state.shared.selected_fitter
    if fitter_id is None or fitter_id is UNDEFINED:
        fitter_id = "amoeba"
    fitter_settings = api.FITTER_DEFAULTS[fitter_id]

    api.state.parallel = options.parallel

    # if args.steps is not None:
    #     fitter_settings["steps"] = args.steps

    if options.filename is not None:
        filepath = Path(options.filename).absolute()
        pathlist = list(filepath.parent.parts)
        filename = filepath.name
        logger.debug(f"fitter for filename {filename} is {fitter_id}")

        async def load_problem(App=None):
            await api.load_problem_file(pathlist, filename)

        app.on_startup.append(load_problem)

    if options.start:

        async def start_fit(App=None):
            if api.state.problem is not None:
                await api.start_fit_thread(fitter_id, fitter_settings["settings"], options.exit)

        app.on_startup.append(start_fit)
    else:
        # signal that no fit is running at startup, even if a fit was
        # interrupted and the state was saved:
        app.on_startup.append(lambda App: api.state.shared.set("active_fit", {}))

    async def notice(message: str):
        logger.info(message)

    app.on_cleanup.append(lambda App: notice("cleanup task"))
    app.on_shutdown.append(lambda App: notice("shutdown task"))
    # not sure why, but have to call shutdown twice to get it to work:
    app.on_shutdown.append(lambda App: api.shutdown())
    app.on_shutdown.append(lambda App: notice("shutdown complete"))

    # set initial path to cwd:
    app.add_routes(routes)
    hostname = "localhost" if not options.external else "0.0.0.0"

    if sock is None:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((hostname, options.port))
    host, port = sock.getsockname()
    api.state.hostname = host
    api.state.port = port
    if options.hub is not None:

        async def register_instance(application: web.Application):
            async with ClientSession() as client_session:
                await client_session.post(options.hub, json={"host": hostname, "port": port})

        app.on_startup.append(register_instance)
    if not options.headless:
        import webbrowser

        async def open_browser(app: web.Application):
            loop = asyncio.get_event_loop()
            loop.call_later(0.5, lambda: webbrowser.open_new_tab(f"http://{hostname}:{port}/"))

        app.on_startup.append(open_browser)

    if options.convergence_heartbeat:
        enable_convergence_kernel_heartbeat()

    if TRACE_MEMORY:
        import tracemalloc

        tracemalloc.start()

    return sock


def main(options: Optional[OPTIONS_CLASS] = None, sock: Optional[socket.socket] = None):
    # this entrypoint will be used to start gui, so set headless = False
    # (other contexts e.g. jupyter notebook will directly call start_app)
    logger.addHandler(console_handler)
    options = get_commandline_options(arg_defaults={"headless": False}) if options is None else options
    logger.info(dict(options=options))
    setup_sio_api()
    runsock = setup_app(options=options, sock=None)
    web.run_app(app, sock=runsock)


async def start_app(
    options: OPTIONS_CLASS = OPTIONS_CLASS(),
    sock: socket.socket = None,
    jupyter_link: bool = False,
    jupyter_heartbeat: bool = False,
):
    # this function is called from jupyter notebook, so set headless = True
    options.headless = True
    # redirect logging to a list
    logger.addHandler(list_handler)
    setup_sio_api()
    runsock = setup_app(options=options, sock=sock)
    runner = web.AppRunner(app, handle_signals=False)
    await runner.setup()
    site = web.SockSite(runner, sock=runsock)
    await site.start()

    if jupyter_heartbeat:
        enable_convergence_kernel_heartbeat()

    if jupyter_link:
        return open_tab_link()
    else:
        url = get_server_url()
        print(f"webserver started: {url}")


def create_server_task():
    return asyncio.create_task(start_app())


def get_server_url():
    from bumps.webview.server import api

    port = getattr(api.state, "port", None)
    if port is None:
        raise ValueError("The web server has not been started.")

    # detect if running through Jupyter Hub
    if "JUPYTERHUB_SERVICE_PREFIX" in os.environ:
        url = f"{os.environ['JUPYTERHUB_SERVICE_PREFIX']}/proxy/{port}/"
    elif api.state.hostname in ("localhost", "127.0.0.1"):  # local server
        url = f"http://{api.state.hostname}:{port}/"
    else:  # external server, e.g. TACC
        url = f"/proxy/{port}/"
    return url


def display_inline_jupyter(width: Union[str, int] = "100%", height: Union[str, int] = 600, single_panel=None) -> None:
    """
    Display the web server in an iframe.

    This is useful for displaying the web server in a Jupyter notebook.

    :param width: The width of the iframe.
    :param height: The height of the iframe.
    """
    from IPython.display import display, IFrame

    url = get_server_url()
    kwargs = dict(single_panel=single_panel) if single_panel is not None else {}
    display(
        IFrame(
            src=url,
            width=width,
            height=height,
            extras=['style="resize: both;"'],
            **kwargs,
        )
    )


def open_tab_link(single_panel=None) -> None:
    """
    Open the web server in a new tab in the default web browser.
    """
    from IPython.display import Javascript, display, HTML

    url = get_server_url()
    if single_panel is not None:
        url += f"?single_panel={single_panel}"
    src = f'<h3><a href="{url}" target="_blank">Open Webview in Tab</a></h3>'
    display(HTML(src))


if __name__ == "__main__":
    main()



"""
Interface compatible with matlab dream.

Usage
-----

This interface is identical to dream in matlab::

    [Sequences, Reduced_Seq, X, output, hist_logp] = \
      dream(MCMCPar, ParRange, Measurement, ModelName, Extra, option)

With care, you will be able to use the same model definition file
on both platforms, with only minor edits required to switch between them.
Clearly, you won't be able to use if statements and loops since the
syntax for python and matlab are incompatible.  Similarly, you will
have to be careful about indentation and line breaks.  And create your
model without comments.

Python requires that structures be defined before you assign values to
their fields, so you will also need the following lines.  Note however,
that they are safe to use in matlab as well::

    MCMCPar = struct()
    Extra = struct()
    Measurement = struct()
    ParRange = struct()


Another challenge is *ModelName*.  In matlab this is the name of the
m-file that contains the model definition.  Following that convention,
we will try to using "from <ModelName> import <ModelName>", and if this
fails, assume that *ModelName* is actually the function itself.  For
this to work you will need to translate the function in ModelName.m
to the equivalent function in ModelName.py.

*option* is the same option number as before

IPython usage
-------------

Within ipython you can interact with your models something like
you do in matlab.  For example::

    $ ipython -pylab
    In [1]: from dream.matlab import *
    In [2]: from dream import *
    In [3]: %run example.m


You can now use various dream visualization tools or use the matlab-like
plotting functions from pylab::

    In [4]: out.state.save('modeloutput')
    In [5]: out.state.plot_state()

Command line usage
------------------

You can also run a suitable m-file example from the command line.  This will
place you at an ipython command line with all the variables from your
m-file available.  For example::

    python -m dream.matlab example.m
    In [1]: out.state.save('modeloutput')
    In [2]: out.state.plot_state()

Script usage
------------

You can create a driver script which calls the m-file example and
uses pylab commands to plot the results.  For example::

    -- example.py --
    #!/usr/bin/env python
    from pylab import *
    from dream.matlab import *
    execfile('example.m')

    from dream import *
    out.state.save('modeloutput')
    plot_state(out.state)

This can be run from the command prompt::

    $ python example.py

"""

__all__ = ["struct", "dream", "setup", "convert_output"]

import numpy as np

from .core import Dream
from .model import Density, LogDensity, Simulation
from .initpop import cov_init, lhs_init
from .crossover import Crossover, AdaptiveCrossover


class struct(object):
    """
    Matlab compatible structure creation.
    """

    def __init__(self, *pairs, **kw):
        for k, v in zip(pairs[::2], pairs[1::2]):
            setattr(self, k, v)
        for k, v in kw.items():
            setattr(self, k, v)

    def __getattr__(self, k):
        return None


def dream(MCMCPar, ParRange, Measurement, ModelName, Extra, option):
    """
    Emulate the matlab dream call.
    """
    dreamer = setup(MCMCPar, ParRange, Measurement, ModelName, Extra, option)
    dreamer.sample()
    return convert_state(dreamer.state)


def setup(MCMCPar, ParRange, Measurement, ModelName, Extra, option):
    """
    Convert matlab dream models to a python Dream object.
    """
    dreamer = Dream()

    # Problem specification
    bounds = ParRange.minn, ParRange.maxn
    dreamer.bounds_style = Extra.BoundHandling
    if ModelName == "Banshp":
        # specific properties of the banana function
        # Extra.imat is computed from cmat
        # MCMCPar.n is implicit in Extra.cmat
        f = Banana(mu=Extra.mu.flatten(), bpar=Extra.bpar, cmat=Extra.cmat)
        option = 4
    else:
        try:
            f = None  # keep lint happy
            # Try matlab style of having the function in the same named file.
            exec("from " + ModelName + " import " + ModelName + " as f")
        except ImportError:
            # The import failed; hope the caller supplied a function instead.
            f = ModelName

    if option == 1:
        model = Density(f, bounds=bounds)
    elif option == 4:
        model = LogDensity(f, bounds=bounds)
    elif option in [2, 3, 5]:
        # Measurement.N is implicit in Measurement.MeasData
        model = Simulation(f, data=Measurement.MeasData, bounds=bounds, sigma=Measurement.Sigma, gamma=MCMCPar.Gamma)
    else:
        raise ValueError("option should be in 1 to 5")
    dreamer.model = model

    # Sampling parameters
    if Extra.save_in_memory == "Yes":
        thinning = 1
    elif Extra.reduced_sample_collection == "Yes":
        thinning = Extra.T
    else:
        thinning = 1
    dreamer.thinning = thinning
    dreamer.draws = MCMCPar.ndraw

    # Outlier detection
    T = MCMCPar.outlierTest
    if T.endswith("_test"):
        T = T[:-5]
    dreamer.outlier_test = T

    # DE parameters
    dreamer.DE_steps = MCMCPar.steps
    dreamer.DE_pairs = MCMCPar.DEpairs
    dreamer.DE_eps = MCMCPar.eps

    # Initial population
    if Extra.InitPopulation == "COV_BASED":
        pop = cov_init(N=MCMCPar.seq, x=Extra.muX.flatten(), cov=Extra.qcov)
    elif Extra.InitPopulation == "LHS_BASED":
        pop = lhs_init(N=MCMCPar.seq, bounds=(ParRange.minn, ParRange.maxn))
    else:
        raise ValueError("Extra.InitPopulation must be COV_BASED or LHS_BASED")
    dreamer.population = pop

    # Crossover parameters
    if Extra.pCR == "Update":
        CR = AdaptiveCrossover(MCMCPar.nCR)
    else:
        CR = Crossover(1.0 / MCMCPar.nCR)
    dreamer.CR = CR

    # Delayed rejection parameters
    dreamer.use_delayed_rejection = Extra.DR == "Yes"
    dreamer.DR_scale = Extra.DRscale

    return dreamer


def convert_state(state):
    """
    Convert a completed dream run into a form compatible with the
    matlab dream interface::

        Sequences, Reduced_Seq, X, out, hist_logp

    The original state is stored in out.state
    """

    _, points, logp = state.sample()
    logp = logp[:, :, None]
    Sequences = np.concatenate((points, np.exp(logp), logp), axis=2)
    X = Sequences[-1, :, :]

    draws, logp = state.logp()
    hist_logp = np.concatenate((draws[:, None], logp), axis=1)

    out = struct()
    draws, AR = state.acceptance_rate()
    out.AR = np.concatenate((draws[:, None], AR[:, None]), axis=1)
    draws, w = state.CR_weight()
    out.CR = np.concatenate((draws[:, None], w), axis=1)
    out.outlier = state.outliers()[:, :2]
    Nupdate, Nvar = len(draws), points.shape[2]
    R = np.zeros((Nupdate, Nvar), "d")  # R is no longer calculated
    out.R_stat = np.concatenate((draws[:, None], R), axis=1)

    # save the dreamer state data structure  as well
    out.state = state

    return Sequences, Sequences, X, out, hist_logp


def run_script(filename):
    exec(compile(open(filename).read(), filename, "exec"))


class Banana(object):
    """
    Banana shaped function.

    Note that this is not one of the N dimensional Rosenbrock variants
    documented on wikipedia as it only operates "banana-like" in
    the x0-x1 plane.
    """

    def __init__(self, mu, bpar, cmat):
        self.mu, self.bpar, self.cmat = mu, bpar, cmat
        self.imat = np.linalg.inv(cmat)

    def __call__(self, x):
        x = x + 0  # make a copy
        x[1] += self.bpar * (x[0] ** 2 - 100)
        ret = -0.5 * np.dot(np.dot(x[None, :], self.imat), x[:, None])
        return ret[0, 0]


def main():
    import sys

    if len(sys.argv) == 2:
        import pylab

        run_script(sys.argv[1])
        user_ns = pylab.__dict__.copy().update(locals())
        import IPython

        IPython.Shell.IPShell(user_ns=user_ns).mainloop()
    else:
        print("usage: python -m dream.matlab model.m")


if __name__ == "__main__":
    main()



import numpy as np

from .. import monitor
from ..plotutil import coordinated_colors
from .plot_view import PlotView


class ConvergenceMonitor(monitor.Monitor):
    """
    Gather statistics about the best, worst, median and +/- 1 interquartile
    range.  This will be the input for the convergence plot.
    """

    def __init__(self):
        self.pop = []

    def config_history(self, history):
        history.requires(population_values=1, value=1)

    def __call__(self, history):
        best = history.value[0]
        try:
            pop = history.population_values[0]
            n = len(pop)
            p = np.sort(pop)
            (
                QI,
                Qmid,
            ) = int(0.2 * n), int(0.5 * n)
            self.pop.append((best, p[0], p[QI], p[Qmid], p[-1 - QI], p[-1]))
        except (AttributeError, TypeError):
            self.pop.append((best,))

    def progress(self):
        if not self.pop:
            return dict(pop=np.empty((0, 1), "d"))
        else:
            return dict(pop=np.array(self.pop))


class ConvergenceView(PlotView):
    title = "Convergence"

    def plot(self):
        if not self.plot_state:
            return
        pop, best = self.plot_state
        with self.pylab_interface as pylab:
            pylab.clf()
            ni, npop = pop.shape
            iternum = np.arange(1, ni + 1)
            tail = int(0.25 * ni)
            c = coordinated_colors(base=(0.4, 0.8, 0.2))
            if npop == 5:
                pylab.fill_between(iternum[tail:], pop[tail:, 1], pop[tail:, 3], color=c["light"], label="_nolegend_")
                pylab.plot(iternum[tail:], pop[tail:, 2], label="80% range", color=c["base"])
                pylab.plot(iternum[tail:], pop[tail:, 0], label="_nolegend_", color=c["base"])
            pylab.plot(iternum[tail:], best[tail:], label="best", color=c["dark"])
            pylab.xlabel("iteration number")
            pylab.ylabel("chisq")
            pylab.legend()
            # pylab.gca().set_yscale('log')
            pylab.draw()

    def update(self, best, pop):
        self.plot_state = pop, best
        self.plot()

    def OnFitProgress(self, event):
        if event.problem != self.model:
            return
        pop = 2 * event.pop / self.model.dof
        self.update(pop[:, 0], pop[:, 1:])



"""
Build trace plots (plotly)
"""

from typing import Union

__all__ = ["plot_trace"]

import numpy as np
from .colors import COLORS


# TODO: when minimum python version is 3.10, can use | to combine types
def plot_trace(x: Union[list, np.ndarray], ys: Union[list, np.ndarray], label="", alpha=0.4):
    import plotly.graph_objs as go

    traces = []
    color_idx = 0
    for y in ys:
        color = COLORS[color_idx % len(COLORS)]
        traces.append(go.Scatter(x=x, y=y, mode="lines", showlegend=False, line=dict(color=color), opacity=alpha))
        color_idx += 1

    fig = go.Figure(data=traces)

    fig.update_layout(
        template="plotly_white",
        xaxis_title=dict(text="Generation number"),
        yaxis_title=dict(text=label),
        yaxis=dict(exponentformat="e"),
        showlegend=False,
    )

    return fig



r"""
Estimate entropy after a fit.

The :func:`gmm_entropy` function computes the entropy from a Gaussian mixture
model. This provides a reasonable estimate even for non-Gaussian distributions.
This is the recommended method for estimating the entropy of a sample.

The :func:`cov_entropy` method computes the entropy associated with the
covariance matrix.  This covariance matrix can be estimated during the
fitting procedure (BFGS updates an estimate of the Hessian matrix for example),
or computed by estimating derivatives when the fit is complete.

The :class:`MVNEntropy` class estimates the covariance from an MCMC sample and
uses this covariance to estimate the entropy.  This gives a better
estimate of the entropy than the equivalent direct calculation, which requires
many more samples for a good kernel density estimate.  The *reject_normal*
attribute is *True* if the MCMC sample is significantly different from normal.
Unfortunately, this almost always the case for any reasonable sample size that
isn't strictly gaussian.

The :func:`entropy` function computes the entropy directly from a set
of MCMC samples, normalized by a scale factor computed from the kernel density
estimate at a subset of the points.\ [#Kramer]_

There are many other entropy calculations implemented within this file, as
well as a number of sampling distributions for which the true entropy is known.
Furthermore, entropy was computed against dream output and checked for
consistency. None of the methods is truly excellent in terms of minimum
sample size, maximum dimensions and speed, but many of them are pretty
good.

The following is an informal summary of the results from different algorithms
applied to DREAM output::

        from .entropy import Timer as T

        # Try MVN ... only good for normal distributions, but very fast
        with T(): M = entropy.MVNEntropy(drawn.points)
        print("Entropy from MVN: %s"%str(M))

        # Try wnn ... no good.
        with T(): S_wnn, Serr_wnn = entropy.wnn_entropy(drawn.points, n_est=20000)
        print("Entropy from wnn: %s"%str(S_wnn))

        # Try wnn with bootstrap ... still no good.
        with T(): S_wnn, Serr_wnn = entropy.wnn_bootstrap(drawn.points)
        print("Entropy from wnn bootstrap: %s"%str(S_wnn))

        # Try wnn entropy with thinning ... still no good.
        #drawn = self.draw(portion=portion, vars=vars,
        #                  selection=selection, thin=10)
        with T(): S_wnn, Serr_wnn = entropy.wnn_entropy(points)
        print("Entropy from wnn: %s"%str(S_wnn))

        # Try wnn with gmm ... still no good
        with T(): S_wnn, Serr_wnn = entropy.wnn_entropy(drawn.points, n_est=20000, gmm=20)
        print("Entropy from wnn with gmm: %s"%str(S_wnn))

        # Try pure gmm ... pretty good
        with T(): S_gmm, Serr_gmm = entropy.gmm_entropy(drawn.points, n_est=10000)
        print("Entropy from gmm: %s"%str(S_gmm))

        # Try kde from statsmodels ... pretty good
        with T(): S_kde_stats = entropy.kde_entropy_statsmodels(drawn.points, n_est=10000)
        print("Entropy from kde statsmodels: %s"%str(S_kde_stats))

        # Try kde from sklearn ... pretty good
        with T(): S_kde = entropy.kde_entropy_sklearn(drawn.points, n_est=10000)
        print("Entropy from kde sklearn: %s"%str(S_kde))

        # Try kde from sklearn at points from gmm ... pretty good
        with T(): S_kde_gmm = entropy.kde_entropy_sklearn_gmm(drawn.points, n_est=10000)
        print("Entropy from kde+gmm: %s"%str(S_kde_gmm))

        # Try Kramer ... pretty good, but doesn't support marginal entropy
        with T(): S, Serr = entropy.entropy(drawn.points, drawn.logp, N_entropy=n_est)
        print("Entropy from Kramer: %s"%str(S))


.. [#Kramer]
    Kramer, A., Hasenauer, J., Allgower, F., Radde, N., 2010.
    Computation of the posterior entropy in a Bayesian framework
    for parameter estimation in biological networks,
    in: 2010 IEEE International Conference on Control Applications (CCA).
    Presented at the 2010 IEEE International Conference on
    Control Applications (CCA), pp. 493-498.
    doi:10.1109/CCA.2010.5611198


.. [#Turjillo-Ortiz]
    Trujillo-Ortiz, A. and R. Hernandez-Walls. (2003). Mskekur: Mardia's
        multivariate skewness and kurtosis coefficients and its hypotheses
        testing. A MATLAB file. [WWW document].
        `<http://www.mathworks.com/matlabcentral/fileexchange/loadFile.do?objectId=3519>`_

.. [#Mardia1970]
    Mardia, K. V. (1970), Measures of multivariate skewnees and kurtosis with
        applications. Biometrika, 57(3):519-530.

.. [#Mardia1974]
    Mardia, K. V. (1974), Applications of some measures of multivariate skewness
        and kurtosis for testing normality and robustness studies. Sankhy A,
        36:115-128

.. [#Stevens]
    Stevens, J. (1992), Applied Multivariate Statistics for Social Sciences.
        2nd. ed. New-Jersey:Lawrance Erlbaum Associates Publishers. pp. 247-248.

"""

__all__ = ["entropy", "gmm_entropy", "cov_entropy", "wnn_entropy", "MVNEntropy"]

import numpy as np
from numpy import e, exp, log, log2, mean, nan, pi, sqrt, std
from numpy.random import choice, permutation
from scipy import stats
from scipy.special import digamma, gammaln
from scipy.stats import chi2, norm

LN2 = log(2)


def standardize(x):
    """
    Standardize the points by removing the mean and scaling by the standard
    deviation.
    """
    # TODO: check if it is better to multiply by inverse covariance
    # That would serve to unrotate and unscale the dimensions together,
    # but squishing them down individually might be just as good.

    # compute zscores for the each variable independently
    mu, sigma = mean(x, axis=0), std(x, axis=0, ddof=1)
    # Protect against NaN when sigma is zero.  If sigma is zero
    # then all points are equal, so x == mu and z-score is zero.
    return (x - mu) / (sigma + (sigma == 0.0)), mu, sigma


def kde_entropy_statsmodels(points, n_est=None):
    """
    Use statsmodels KDEMultivariate pdf to estimate entropy.

    Density evaluated at sample points.

    Slow and fails for bimodal, dirichlet; poor for high dimensional MVN.
    """
    from statsmodels.nonparametric.kernel_density import KDEMultivariate

    n, d = points.shape

    # Default to the full set
    if n_est is None:
        n_est = n

    # reduce size of draw to n_est
    if n_est >= n:
        x = points
    else:
        x = points[permutation(n)[:n_est]]
        n = n_est

    predictor = KDEMultivariate(data=x, var_type="c" * d)
    p = predictor.pdf()
    H = -np.mean(log(p))
    return H / LN2


def kde_entropy_sklearn(points, n_est=None):
    """
    Use sklearn.neigbors.KernelDensity pdf to estimate entropy.

    Data is standardized before analysis.

    Sample points drawn from the kernel density estimate.

    Fails for bimodal and dirichlet, similar to statsmodels kde.
    """
    n, d = points.shape

    # Default to the full set
    if n_est is None:
        n_est = n

    # reduce size of draw to n_est
    if n_est >= n:
        x = points
    else:
        x = points[permutation(n)[:n_est]]
        n = n_est

    # logp = sklearn_log_density(points, evaluation_points=n_est)
    logp = sklearn_log_density(x, evaluation_points=x)
    H = -np.mean(logp)
    return H / LN2


def kde_entropy_sklearn_gmm(points, n_est=None, n_components=None):
    """
    Use sklearn.neigbors.KernelDensity pdf to estimate entropy.

    Data is standardized before kde.

    Sample points drawn from gaussian mixture model from original points.

    Fails for bimodal and dirichlet, similar to statsmodels kde.
    """
    from sklearn.mixture import BayesianGaussianMixture as GMM

    n, d = points.shape

    # Default to the full set
    if n_est is None:
        n_est = n

    # reduce size of draw to n_est
    if n_est >= n:
        x = points
    else:
        x = points[permutation(n)[:n_est]]
        n = n_est

    if n_components is None:
        n_components = int(5 * sqrt(d))

    predictor = GMM(
        n_components=n_components,
        covariance_type="full",
        # verbose=True,
        max_iter=1000,
    )
    predictor.fit(x)
    evaluation_points, _ = predictor.sample(n_est)

    logp = sklearn_log_density(x, evaluation_points=evaluation_points)
    H = -np.mean(logp)
    return H / LN2


def gmm_entropy(points, n_est=None, n_components=None):
    r"""
    Use sklearn.mixture.BayesianGaussianMixture to estimate entropy.

    *points* are the data points in the sample.

    *n_est* are the number of points to use in the estimation; default is
    10,000 points, or 0 for all the points.

    *n_components* are the number of Gaussians in the mixture. Default is
    $5 \sqrt{d}$ where $d$ is the number of dimensions.

    Returns estimated entropy and uncertainty in the estimate.

    This method uses BayesianGaussianMixture from scikit-learn to build a
    model of the point distribution, then uses Monte Carlo sampling to
    determine the entropy of that distribution. The entropy uncertainty is
    computed from the variance in the MC sample scaled by the number of
    samples. This does not incorporate any uncertainty in the sampling that
    generated the point distribution or the uncertainty in the GMM used to
    model that distribution.
    """
    # from sklearn.mixture import GaussianMixture as GMM
    from sklearn.mixture import BayesianGaussianMixture as GMM

    n, d = points.shape

    # Default to the full set
    if n_est is None:
        n_est = 10000
    elif n_est == 0:
        n_est = n

    # reduce size of draw to n_est
    if n_est >= n:
        x = points
        n_est = n
    else:
        x = points[permutation(n)[:n_est]]
        n = n_est

    if n_components is None:
        n_components = int(5 * sqrt(d))

    ## Standardization doesn't seem to help
    ## Note: sigma may be zero
    # x, mu, sigma = standardize(x)   # if standardized
    predictor = GMM(
        n_components=n_components,
        covariance_type="full",
        # verbose=True,
        max_iter=1000,
    )
    predictor.fit(x)
    eval_x, _ = predictor.sample(n_est)
    weight_x = predictor.score_samples(eval_x)
    H = -np.mean(weight_x)
    # with np.errstate(divide='ignore'): H = H + np.sum(np.log(sigma))   # if standardized
    dH = np.std(weight_x, ddof=1) / sqrt(n)
    ## cross-check against own calcs
    # alt = GaussianMixture(predictor.weights_, mu=predictor.means_, sigma=predictor.covariances_)
    # print("alt", H, alt.entropy())
    # print(np.vstack((weight_x[:10], alt.logpdf(eval_x[:10]))).T)
    return H / LN2, dH / LN2


def wnn_bootstrap(points, k=None, weights=True, n_est=None, reps=10, parts=10):
    # raise NotImplementedError("deprecated; bootstrap doesn't help.")
    n, d = points.shape
    if n_est is None:
        n_est = n // parts

    results = [wnn_entropy(points, k=k, weights=weights, n_est=n_est) for _ in range(reps)]
    # print(results)
    S, Serr = list(zip(*results))
    return np.mean(S), np.std(S)


def wnn_entropy(points, k=None, weights=True, n_est=None, gmm=None):
    r"""
    Weighted Kozachenko-Leonenko nearest-neighbour entropy calculation.

    *k* is the number of neighbours to consider, with default $k=n^{1/3}$

    *n_est* is the number of points to use for estimating the entropy,
    with default $n_\rm{est} = n$

    *weights* is True for default weights, False for unweighted (using the
    distance to the kth neighbour only), or a vector of weights of length *k*.

    *gmm* is the number of gaussians to use to model the distribution using
    a gaussian mixture model.  Default is 0, and the points represent an
    empirical distribution.

    Returns entropy H in bits and its uncertainty.

    Berrett, T. B., Samworth, R.J., Yuan, M., 2016. Efficient multivariate
    entropy estimation via k-nearest neighbour distances.
    DOI:10.1214/18-AOS1688 https://arxiv.org/abs/1606.00304
    """
    from sklearn.neighbors import NearestNeighbors

    n, d = points.shape

    # Default to the full set
    if n_est is None:
        n_est = 10000
    elif n_est == 0:
        n_est = n

    # reduce size of draw to n_est
    if n_est >= n:
        x = points
        n_est = n
    else:
        x = points[permutation(n)[:n_est]]
        n = n_est

    # Default k based on n
    if k is None:
        # Private communication: cube root of n is a good choice for k
        # Personal observation: k should be much bigger than d
        k = max(int(n ** (1 / 3)), 3 * d)

    # If weights are given then use them (setting the appropriate k),
    # otherwise use the default weights.
    if isinstance(weights, bool):
        weights = _wnn_weights(k, d, weights)
    else:
        k = len(weights)
    # print("weights", weights, sum(weights))

    # select knn algorithm
    algorithm = "auto"
    # algorithm = 'kd_tree'
    # algorithm = 'ball_tree'
    # algorithm = 'brute'

    n_components = 0 if gmm is None else gmm

    # H = 1/n sum_i=1^n sum_j=1^k w_j log E_{j,i}
    # E_{j,i} = e^-Psi(j) V_d (n-1) z_{j,i}^d = C z^d
    # logC = -Psi(j) + log(V_d) + log(n-1)
    # H = 1/n sum sum w_j logC + d/n sum sum w_j log(z)
    #   = sum w_j logC + d/n sum sum w_j log(z)
    #   = A + d/n B
    # H^2 = 1/n sum
    Psi = digamma(np.arange(1, k + 1))
    logVd = d / 2 * log(pi) - gammaln(1 + d / 2)
    logC = -Psi + logVd + log(n - 1)

    # TODO: standardizing points doesn't work.
    # Standardize the data so that distances conform.  This is equivalent to
    # a u-substitution u = sigma x + mu, so the integral needs to be corrected
    # for dU = det(sigma) dx.  Since the standardization squishes the dimensions
    # independently, sigma is a diagonal matrix, with the determinant equal to
    # the product of the diagonal elements.
    # x, mu, sigma = standardize(x)  # Note: sigma may be zero
    # detDU = np.prod(sigma)
    detDU = 1.0

    if n_components > 0:
        # Use Gaussian mixture to model the distribution
        from sklearn.mixture import GaussianMixture as GMM

        predictor = GMM(n_components=gmm, covariance_type="full")
        predictor.fit(x)
        eval_x, _ = predictor.sample(n_est)
        # weight_x = predictor.score_samples(eval_x)
        skip = 0
    else:
        # Empirical distribution
        # TODO: should we use the full draw for kNN and a subset for eval points?
        # Choose a subset for evaluating the entropy estimate, if desired
        # print(n_est, n)
        # eval_x = x if n_est >= n else x[permutation(n)[:n_est]]
        eval_x = x
        # weight_x = 1
        skip = 1

    tree = NearestNeighbors(algorithm=algorithm, n_neighbors=k + skip)
    tree.fit(x)
    dist, _ind = tree.kneighbors(eval_x, n_neighbors=k + skip, return_distance=True)
    # Remove first column. Since test points are in x, the first column will
    # be a point from x with distance 0, and can be ignored.
    if skip:
        dist = dist[:, skip:]
    # Find log distances.  This can be problematic for MCMC runs where a
    # step is rejected, and therefore identical points are in the distribution.
    # Ignore them by replacing these points with nan and using nanmean.
    # TODO: need proper analysis of duplicated points in MCMC chain
    dist[dist == 0] = nan
    logdist = log(dist)
    H_unweighted = logC + d * np.nanmean(logdist, axis=0)
    H = np.dot(H_unweighted, weights)[0]
    Hsq_k = np.nanmean((logC[-1] + d * logdist[:, -1]) ** 2)
    # TODO: abs shouldn't be needed?
    if Hsq_k < H**2:
        print("warning: avg(H^2) < avg(H)^2")
    dH = sqrt(abs(Hsq_k - H**2) / n_est)
    # print("unweighted", H_unweighted)
    # print("weighted", H, Hsq_k, H**2, dH, detDU, LN2)
    return H * detDU / LN2, dH * detDU / LN2


def _wnn_weights(k, d, weighted=True):
    # Private communication: ignore w_j = 0 constraints (they are in the
    # paper for mathematical nicety), and find the L2 norm of the
    # remaining underdeterimined system described in Eq 2.
    # Personal observation: k should be some small multiple of d
    # otherwise the weights blow up.
    if d < 4 or not weighted:
        # with few dimensions go unweighted with the kth nearest neighbour.
        return np.array([[0.0] * (k - 1) + [1.0]]).T
    j = np.arange(1, k + 1)
    sum_zero = [exp(gammaln(j + 2 * i / d) - gammaln(j)) for i in range(1, d // 4 + 1)]
    sum_one = [[1.0] * k]
    A = np.array(sum_zero + sum_one)
    b = np.array([[0.0] * (d // 4) + [1.0]]).T
    return np.dot(np.linalg.pinv(A), b)


def scipy_stats_density(sample_points, evaluation_points):  # pragma: no cover
    """
    Estimate the probability density function from which a set of sample
    points was drawn and return the estimated density at the evaluation points.
    """
    ## standardize data so that we can use uniform bandwidth
    ## Note: this didn't help with singular matrix
    ## Note: if re-enable, protect against sigma=0 in some dimensions
    # mu, sigma = mean(data, axis=0), std(data, axis=0)
    # data,points = (data - mu)/sigma, (points - mu)/sigma

    kde = stats.gaussian_kde(sample_points)
    return kde(evaluation_points)


def sklearn_log_density(sample_points, evaluation_points):
    """
    Estimate the log probability density function from which a set of sample
    points was drawn and return the estimated density at the evaluation points.

    *sample_points* is an [n x m] matrix.

    *evaluation_points* is the set of points at which to evaluate the kde.

    Note: if any dimension has all points equal then the entire distribution
    is treated as a dirac distribution with infinite density at each point.
    This makes the entropy calculation better behaved (narrowing the
    distribution increases the entropy) but is not so useful in other contexts.
    Other packages will (correctly) ignore dimensions of width zero.
    """
    # Ugly hack warning: if *evaluation_points* is an integer, then sample
    # that many points from the kde and return the log density at each
    # sampled point.  Since the code that uses this is looking only at
    # the mean log density, it doesn't need the sample points themselves.
    # This interface should be considered internal to the entropy module
    # and not used by outside functions.  If you need it externally, then
    # restructure the api so that the function always returns both the
    # points and the density, as well as any other function (such as the
    # denisty function and the sister function scipy_stats_density) so
    # that all share the new interface.

    from sklearn.neighbors import KernelDensity

    # Standardize data so we can use spherical kernels and uniform bandwidth
    data, mu, sigma = standardize(sample_points)

    # Note that sigma will be zero for dimensions w_o where all points are equal.
    # With P(w) = P(w, w_o) / P(w_o | w) and P(w_o) = 1 for all points in
    # the set, then P(w) = P(w, w_o) and we can ignore the zero dimensions.
    # However, as another ugly hack, we want the differential entropy to go
    # to -inf as the distribution narrows, so pretend that P = 0 everywhere.
    # Uncomment the following line to return the sample probability instead.
    ## sigma[sigma == 0.] = 1.

    # Silverman bandwidth estimator
    n, d = sample_points.shape
    bandwidth = (n * (d + 2) / 4.0) ** (-1.0 / (d + 4))

    # print("starting grid search for bandwidth over %d points"%n)
    # from sklearn.grid_search import GridSearchCV
    # from numpy import logspace
    # params = {'bandwidth': logspace(-1, 1, 20)}
    # fitter = GridSearchCV(KernelDensity(), params)
    # fitter.fit(data)
    # kde = fitter.best_estimator_
    # print("best bandwidth: {0}".format(kde.bandwidth))
    # import time; T0 = time.time()
    kde = KernelDensity(kernel="gaussian", bandwidth=bandwidth, rtol=1e-6, atol=1e-6)
    kde.fit(data)

    if isinstance(evaluation_points, int):
        # For generated points, they already follow the distribution
        points = kde.sample(n)
    elif evaluation_points is not None:
        # Standardized evaluation points to match sample distribution
        # Note: for dimensions where all sample points are equal, sigma
        # has been artificially set equal to one.  This means that the
        # evaluation points which do not match the sample value will
        # use the simple differences for the z-score rather than
        # pushing them out to plus/minus infinity.
        points = (evaluation_points - mu) / (sigma + (sigma == 0.0))
    else:
        points = sample_points

    # Evaluate pdf, scaling the resulting density by sigma to correct the area.
    # If sigma is zero, return entropy as -inf;  this seems to not be the
    # case for discrete distributions (consider Bernoulli with p=1, q=0,
    #  => H = -p log p - q log q = 0), so need to do something else, both
    # for the kde and for the entropy calculation.
    with np.errstate(divide="ignore"):
        log_pdf = kde.score_samples(points) - np.sum(np.log(sigma))

    return log_pdf


def sklearn_density(sample_points, evaluation_points):
    """
    Estimate the probability density function from which a set of sample
    points was drawn and return the estimated density at the evaluation points.
    """
    return exp(sklearn_log_density(sample_points, evaluation_points))


# scipy kde fails with singular matrix, so we will use scikit.learn
# density = scipy_stats_density
density = sklearn_density


def entropy(points, logp, N_entropy=10000, N_norm=2500):
    r"""
    Return entropy estimate and uncertainty from a random sample.

    *points* is a set of draws from an underlying distribution, as returned
    by a Markov chain Monte Carlo process for example.

    *logp* is the log-likelihood for each draw.

    *N_norm* is the number of points $k$ to use to estimate the posterior
    density normalization factor $P(D) = \hat N$, converting
    from $\log( P(D|M) P(M) )$ to $\log( P(D|M)P(M)/P(D) )$. The relative
    uncertainty $\Delta\hat S/\hat S$ scales with $\sqrt{k}$, with the
    default *N_norm=2500* corresponding to 2% relative uncertainty.
    Computation cost is $O(nk)$ where $n$ is number of points in the draw.

    *N_entropy* is the number of points used to estimate the entropy
    $\hat S = - \int P(M|D) \log P(M|D)$ from the normalized log likelihood
    values.
    """

    # Use a random subset to estimate density
    if N_norm >= len(logp):
        norm_points = points
    else:
        idx = permutation(len(points))[:N_entropy]
        norm_points = points[idx]

    # Use a different subset to estimate the scale factor between density
    # and logp.
    if N_entropy is None:
        N_entropy = 10000
    if N_entropy >= len(logp):
        entropy_points, eval_logp = points, logp
    else:
        idx = permutation(len(points))[:N_entropy]
        entropy_points, eval_logp = points[idx], logp[idx]

    """
    # Try again, just using the points from the high probability regions
    # to determine the scale factor
    N_norm = min(len(logp), 5000)
    N_entropy = int(0.8*N_norm)
    idx = np.argsort(logp)
    norm_points = points[idx[-N_norm:]]
    entropy_points = points[idx[-N_entropy:]]
    eval_logp = logp[idx[-N_entropy:]]
    """

    # Normalize p to a peak probability of 1 so that exp() doesn't underflow.
    #
    # This should be okay since for the normalizing constant C:
    #
    #      u' = e^(ln u + ln C) = e^(ln u)e^(ln C) = C u
    #
    # Using eq. 11 of Kramer with u' substituted for u:
    #
    #      N_est = < u'/p > = < C u/p > = C < u/p >
    #
    #      S_est = - < ln q >
    #            = - < ln (u'/N_est) >
    #            = - < ln C + ln u - ln (C <u/p>) >
    #            = - < ln u + ln C - ln C - ln <u/p> >
    #            = - < ln u - ln <u/p> >
    #            = - < ln u > + ln <u/p>
    #
    # Uncertainty comes from eq. 13:
    #
    #      N_err^2 = 1/(k-1) sum( (u'/p - <u'/p>)^2 )
    #              = 1/(k-1) sum( (C u/p - <C u/p>)^2 )
    #              = C^2 std(u/p)^2
    #      S_err = std(u'/p) / <u'/p> = (C std(u/p))/(C <u/p>) = std(u/p)/<u/p>
    #
    # So even though the constant C shows up in N_est, N_err, it cancels
    # again when S_est, S_err is formed.
    log_scale = np.max(eval_logp)
    # print("max log sample: %g"%log_scale)
    eval_logp -= log_scale

    # Compute entropy and uncertainty in nats
    # Note: if all values are the same in any dimension then we have a dirac
    # functional with infinite probability at every sample point, and the
    # differential entropy estimate will yield H = -inf.
    rho = density(norm_points, entropy_points)
    # print(rho.min(), rho.max(), eval_logp.min(), eval_logp.max())
    frac = exp(eval_logp) / rho
    n_est, n_err = mean(frac), std(frac)
    if n_est == 0.0:
        s_est, s_err = -np.inf, 0.0
    else:
        s_est = log(n_est) - mean(eval_logp)
        s_err = n_err / n_est
    # print(n_est, n_err, s_est/LN2, s_err/LN2)
    ##print(np.median(frac), log(np.median(frac))/LN2, log(n_est)/LN2)
    if False:
        import pylab

        idx = pylab.argsort(entropy_points[:, 0])
        pylab.figure()
        pylab.subplot(221)
        pylab.hist(points[:, 0], bins=50, density=True, log=True)
        pylab.plot(entropy_points[idx, 0], rho[idx], label="density")
        pylab.plot(entropy_points[idx, 0], exp(eval_logp + log_scale)[idx], label="p")
        pylab.ylabel("p(x)")
        pylab.legend()
        pylab.subplot(222)
        pylab.hist(points[:, 0], bins=50, density=True, log=False)
        pylab.plot(entropy_points[idx, 0], rho[idx], label="density")
        pylab.plot(entropy_points[idx, 0], exp(eval_logp + log_scale)[idx], label="p")
        pylab.ylabel("p(x)")
        pylab.legend()
        pylab.subplot(212)
        pylab.plot(entropy_points[idx, 0], frac[idx], ".")
        pylab.xlabel("P[0] value")
        pylab.ylabel("p(x)/kernel density")

    # return entropy and uncertainty in bits
    return s_est / LN2, s_err / LN2


class MVNEntropy(object):
    """
    Multivariate normal entropy approximation.

    Uses Mardia's multivariate skewness and kurtosis test to estimate normality.

    *x* is a set of points

    *alpha* is the cutoff for the normality test.

    *max_points* is the maximum number of points to use when checking
    normality.  Since the normality test is $O(n^2)$ in memory and time,
    where $n$ is the number of points, *max_points* defaults to 1000. The
    entropy is computed from the full dataset.

    The returned object has the following attributes:

        *p_kurtosis* is the p-value for the kurtosis normality test

        *p_skewness* is the p-value for the skewness normality test

        *reject_normal* is True if either the the kurtosis or the skew test
        fails

        *entropy* is the estimated entropy of the best normal approximation
        to the distribution

    """

    # TODO: use robust covariance estimator for mean and covariance
    # FastMSD is available in sklearn.covariance.MinDetCov. There are methods
    # such as (Zhahg, 2012), which may be faster if performance is an issue.
    # [1] Zhang (2012) DOI: 10.5539/ijsp.v1n2p119
    def __init__(self, x, alpha=0.05, max_points=1000):
        # compute Mardia test coefficient
        n, p = x.shape  # num points, num dimensions
        mu = np.mean(x, axis=0)
        C = np.cov(x.T, bias=1) if p > 1 else np.array([[np.var(x.T, ddof=1)]])
        # squared Mahalanobis distance matrix
        # Note: this forms a full n x n matrix of distances, so will
        # fail for a large number of points.  Kurtosis only requires
        # the diagonal elements so can be computed cheaply.  If there
        # is no order to the points, skew could be estimated using only
        # the block diagonal
        dx = (x - mu[None, :])[:max_points]
        D = np.dot(dx, np.linalg.solve(C, dx.T))
        kurtosis = np.sum(np.diag(D) ** 2) / n
        skewness = np.sum(D**3) / n**2

        kurtosis_stat = (kurtosis - p * (p + 2)) / sqrt(8 * p * (p + 2) / n)
        raw_skewness_stat = n * skewness / 6
        # Small sample correction converges to 1 as n increases, so it is
        # always safe to apply it
        small_sample_correction = (p + 1) * (n + 1) * (n + 3) / ((p + 1) * (n + 1) * n - n * 6)
        skewness_stat = raw_skewness_stat * small_sample_correction
        dof = (p * (p + 1) * (p + 2)) / 6  # degrees of freedom for chisq test

        self.p_kurtosis = 2 * (1 - norm.cdf(abs(kurtosis_stat)))
        self.p_skewness = 1 - chi2.cdf(skewness_stat, dof)
        self.reject_normal = self.p_kurtosis < alpha or self.p_skewness < alpha
        # print("kurtosis", kurtosis, kurtosis_stat, self.p_kurtosis)
        # print("skewness", skewness, skewness_stat, self.p_skewness)
        # compute entropy
        self.entropy = cov_entropy(C)

    def __str__(self):
        return "H=%.1f bits%s" % (self.entropy, " (not normal)" if self.reject_normal else "")


def cov_entropy(C):
    """
    Entropy estimate from covariance matrix C
    """
    return 0.5 * (len(C) * log2(2 * pi * e) + log2(abs(np.linalg.det(C))))


def mvn_entropy_bootstrap(points, samples=50):
    """
    Use bootstrap method to estimate entropy and its uncertainty
    """
    n, d = points.shape

    results = []
    for _ in range(samples):
        # sample n points with replacement in 0 ... n-1.
        x = points[choice(n, size=n)]
        C = np.cov(x.T, bias=1) if d > 1 else np.array([[np.var(x.T, ddof=1)]])
        # print(f"cov {samples}, {x.shape}, {C.shape}")
        results.append(cov_entropy(C))

    return np.mean(results), np.std(results)


# ======================================================================
# Testing code
# ======================================================================

# Based on: Eli Bendersky https://stackoverflow.com/a/5849861
# Extended with tic/toc by Paul Kienzle
import time


class Timer(object):
    @staticmethod
    def tic(name=None):
        return Timer(name).toc

    def __init__(self, name=None):
        self.name = name
        self.step_number = 0
        self.tlast = self.tstart = time.time()

    def toc(self, step=None):
        self.step_number += 1
        if step is None:
            step = str(self.step_number)
        label = self.name + "-" + step if self.name else step
        tnext = time.time()
        total = tnext - self.tstart
        delta = tnext - self.tlast
        print("[%s] Elapsed: %s, Delta: %s" % (label, total, delta))
        self.tlast = tnext

    def __enter__(self):
        self.tlast = self.tstart = time.time()

    def __exit__(self, type, value, traceback):
        if self.name:
            print("[%s]" % self.name, end="")
        print("Elapsed: %s" % (time.time() - self.tstart))


def entropy_mc(D, N=1000000):
    logp = D.logpdf(D.rvs(N))
    return -np.mean(logp)
    # return -np.mean(logp[np.isfinite(logp)])


# CRUFT: dirichlet needs transpose of theta for logpdf
class Dirichlet:
    def __init__(self, alpha):
        self.alpha = alpha
        self._dist = stats.dirichlet(alpha)
        self.dim = len(alpha)

    def logpdf(self, theta):
        return self._dist.logpdf(theta.T)

    def rvs(self, *args, **kw):
        x = self._dist.rvs(*args, **kw)
        # Dirichlet logpdf is failing if x=0 for any x when alpha<1.
        # The simplex check allows fudge of 1e-10.
        x[x == 0] = 1e-100
        return x

    def entropy(self, *args, **kw):
        return self._dist.entropy(*args, **kw)


class Box:
    def __init__(self, width=None, center=None):
        if width is None:
            width = np.ones(len(center), dtype="d")
        if center is None:
            center = np.zeros(len(width), dtype="d")
        self.center = center
        self.width = width
        self.dim = len(width)
        self._logpdf = -np.sum(np.log(self.width))

    def rvs(self, size=1):
        x = np.random.rand(size, len(self.width))
        x = (x - 0.5) * self.width + self.center
        return x

    def logpdf(self, theta):
        y = (theta - self.center) / self.width + 0.5
        logp = np.ones(len(theta)) * self._logpdf
        logp[np.any(y < 0, axis=1)] = -np.inf
        logp[np.any(y > 1, axis=1)] = -np.inf
        return logp

    def entropy(self):
        return -self._logpdf


# CRUFT: scipy MVN gives wrong entropy for singular (and near singular) matrices
# This solution gives wrong results for near-singular rvs(), but for the simple
# case of a diagonal Sigma with one zero it does what I need for the test.
class MVNSingular:
    def __init__(self, *args, **kw):
        kw["allow_singular"] = True
        self.dist = stats.multivariate_normal(*args, **kw)

    @property
    def dim(self):
        return self.dist.dim

    def pdf(self, theta):
        return self.dist.pdf(theta)

    def logpdf(self, theta):
        return self.dist.logpdf(theta)

    def rvs(self, size=1):
        return self.dist.rvs(size=size)

    def entropy(self, N=10000):
        # CRUFT scipy==1.10.0: scipy.stats briefly removed the dist.cov attribute.
        if hasattr(self.dist, "cov"):
            cov = self.dist.cov
        else:
            cov = self.dist.cov_object.covariance

        with np.errstate(divide="ignore"):
            return 0.5 * log(np.linalg.det((2 * pi * np.e) * cov))


class GaussianMixture:
    def __init__(self, w, mu=None, sigma=None):
        mu = np.asarray(mu)
        dim = mu.shape[1]
        if sigma is None:
            sigma = [None] * len(mu)
        sigma = [(np.ones(dim) if s is None else np.asarray(s)) for s in sigma]
        sigma = [(np.diag(s) if len(s.shape) == 1 else s) for s in sigma]
        self.dim = dim
        self.weight = np.asarray(w, "d") / np.sum(w)
        self.dist = [stats.multivariate_normal(mean=m, cov=s) for m, s in zip(mu, sigma)]

    def pdf(self, theta):
        return sum(w * D.pdf(theta) for w, D in zip(self.weight, self.dist))

    def logpdf(self, theta):
        return log(self.pdf(theta))

    def rvs(self, size=1):
        # TODO: should randomize the output
        sizes = partition(size, self.weight)
        draws = [D.rvs(size=n) for n, D in zip(sizes, self.dist)]
        return np.random.permutation(np.vstack(draws))

    def entropy(self, N=10000):
        # No analytic expression, so estimate entropy using MC integration
        return entropy_mc(self, N=N)


class MultivariateT:
    def __init__(self, mu=None, sigma=None, df=None):
        if sigma is not None:
            sigma = np.asarray(sigma)
        self.mu = np.zeros(sigma.shape[0]) if mu is None else np.asarray(mu)
        if sigma is None:
            sigma = np.ones(len(mu))
        if len(sigma.shape) == 1:
            sigma = np.diag(sigma)
        self.dim = len(self.mu)
        self.sigma = sigma
        self.df = df
        # Use scipy stats to compute |Sigma| and (x-mu)^T Sigma^{-1} (x - mu),
        # and to estimate dimension p from rank.  Formula for pdf from wikipedia
        # https://en.wikipedia.org/wiki/Multivariate_t-distribution
        from scipy.stats._multivariate import _PSD

        self._psd = _PSD(self.sigma)
        nu, p = self.df, self._psd.rank
        self._log_norm = gammaln((nu + p) / 2) - gammaln(nu / 2) - p / 2 * log(pi * nu) - self._psd.log_pdet / 2

    def logpdf(self, theta):
        dev = theta - self.mu
        maha = np.sum(np.square(np.dot(dev, self._psd.U)), axis=-1)
        nu, p = self.df, self._psd.rank
        return self._log_norm - (nu + p) / 2 * np.log1p(maha / nu)

    def pdf(self, theta):
        return exp(self.logpdf(theta))

    def rvs(self, size=1):
        # From farhawa on stack overflow
        # https://stackoverflow.com/questions/29798795/multivariate-student-t-distribution-with-python
        nu, p = self.df, len(self.mu)
        g = np.tile(np.random.gamma(nu / 2, 2 / nu, size=size), (p, 1)).T
        Z = np.random.multivariate_normal(np.zeros(p), self.sigma, size=size)
        return self.mu + Z / np.sqrt(g)

    def entropy(self, N=100000):
        # No analytic expression, so estimate entropy using MC integration
        return entropy_mc(self, N=N)


def MultivariateCauchy(mu=None, sigma=None):
    return MultivariateT(mu=mu, sigma=sigma, df=1)


class Joint:
    def __init__(self, distributions):
        # Note: list(x) converts any sequence, including generators, into a list
        self.distributions = list(distributions)
        self.dim = len(self.distributions)

    def rvs(self, size=1):
        return np.stack([D.rvs(size=size) for D in self.distributions], axis=-1)

    def pdf(self, theta):
        return exp(self.logpdf(theta))

    def logpdf(self, theta):
        return sum(D.logpdf(theta[..., k]) for k, D in enumerate(self.distributions))

    def cdf(self, theta):
        return exp(self.logcdf(theta))

    def logcdf(self, theta):
        return sum(D.logcdf(theta[..., k]) for k, D in enumerate(self.distributions))

    def sf(self, theta):
        return -np.expm1(self.logcdf(theta))

    def logsf(self, theta):
        return log(self.sf(theta))

    def entropy(self):
        return sum(D.entropy() for D in self.distributions)


def partition(n, w):
    # TODO: build an efficient algorithm for splitting n things into k buckets
    indices = np.arange(len(w), dtype="i")
    choices = np.random.choice(indices, size=n, replace=True, p=w)
    bins = np.arange(len(w) + 1, dtype="f") - 0.5
    sizes, _ = np.histogram(choices, bins=bins)
    return sizes


def _check_entropy(name, D, seed=1, N=10000, N_entropy=None, N_norm=2500, demo=False):
    """
    Check if entropy from a random draw matches analytic entropy.
    """
    use_kramer = use_mvn = use_wnn = use_gmm = use_kde = False
    if demo:
        # use_kramer = True
        # use_wnn = True
        use_mvn = True
        use_gmm = True
        use_kde = True
    else:
        use_kramer = True

    state = np.random.get_state()
    np.random.seed(seed)
    try:
        theta = D.rvs(size=N)
        if getattr(D, "dim", 1) == 1:
            theta = theta.reshape(N, 1)
        if use_kramer:
            logp_theta = D.logpdf(theta)
            logp_theta += 27  # result should be independent of scale factor
            S, Serr = entropy(theta, logp_theta, N_entropy=N_entropy, N_norm=N_norm)
        if use_wnn:
            S_wnn, Serr_wnn = wnn_entropy(theta, n_est=N_entropy)
        if use_gmm:
            S_gmm, Serr_gmm = gmm_entropy(theta, n_est=N_entropy)
        if use_mvn:
            M = MVNEntropy(theta)
            S_mvn = M.entropy
        if use_kde:
            S_kde = kde_entropy_statsmodels(theta, n_est=N_entropy)
            # S_kde = kde_entropy_sklearn(theta, n_est=N_entropy)
            # S_kde = kde_entropy_sklearn_gmm(theta, n_est=N_entropy)
    finally:
        np.random.set_state(state)
    H = D.entropy() / LN2
    if demo:
        print("entropy", N, "~", name, H, end="")
        if use_kramer:
            print(" Kramer", S, Serr, end="")
        if use_wnn:
            print(" wnn", S_wnn, Serr_wnn, end="")
        if use_gmm:
            print(" gmm", S_gmm, Serr_gmm, end="")
        if use_mvn:
            print(" MVN", S_mvn, end="")
        if use_kde:
            print(" KDE", S_kde, end="")
        print()
    else:
        if use_kramer:
            # assert Serr < 0.05*S, "incorrect error est. for Kramer"
            if np.isfinite(H):
                assert abs(S - H) < 3 * Serr, "incorrect est. for Kramer"
            else:
                assert not np.isfinite(S), "incorrect est. for Kramer"
        if use_wnn:
            if np.isfinite(H):
                assert Serr_wnn < 0.05 * S_wnn, "incorrect error est. for wnn"
                assert abs(S_wnn - H) < 3 * Serr_wnn, "incorrect est. for wnn"
            else:
                assert not np.isfinite(S_wnn), "incorrect est. for wnn"


def _show_entropy(name, D, **kw):
    with Timer():
        return _check_entropy(name, D, seed=None, demo=True, **kw)


def _check_smoke(D):
    theta = D.rvs(size=1000)
    if getattr(D, "dim", 1) == 1:
        theta = theta.reshape(-1, 1)
    logp_theta = D.logpdf(theta)
    entropy(theta, logp_theta)
    wnn_entropy(theta)
    MVNEntropy(theta).entropy


def test():
    """check entropy estimates from known distributions"""
    # entropy test is optional: don't test if sklearn is not installed
    try:
        import sklearn
    except ImportError:
        return

    # Smoke test - do all the methods run in 1-D and 10-D?
    _check_smoke(stats.norm(10, 8))
    if hasattr(stats, "multivariate_normal"):
        _check_smoke(stats.multivariate_normal(cov=np.diag([1] * 10)))

    D = stats.norm(10, 8)
    _check_entropy("N[100,8]", D, N=2000)
    _check_entropy("N[100,8]", D, N=12000)
    if hasattr(stats, "multivariate_normal"):
        D = stats.multivariate_normal(cov=np.diag([1, 12**2, 0.2**2]))
        _check_entropy("MVN[1,12,0.2]", D)
        D = stats.multivariate_normal(cov=np.diag([1] * 10))
        _check_entropy("MVN[1]*10", D, N=10000)
        # Make sure zero-width dimensions return H = -inf
        D = MVNSingular(cov=np.diag([1, 1, 0]))
        _check_entropy("MVN[1,1,0]", D, N=10000)
    # raise TestFailure("make bumps testing fail so we know that test harness works")


def mvn_entropy_test():
    """
    Test against results from the R MVN pacakge (using the web version)
    and the matlab Mskekur program (using Octave), both of which produce
    the same value.  Note that MVNEntropy uses the small sample correction
    for the skewness stat since it converges to the large sample value for
    large n.
    """
    x = np.array(
        [
            [2.4, 2.1, 2.4],
            [4.5, 4.9, 5.7],
            [3.5, 1.8, 3.9],
            [3.9, 4.7, 4.7],
            [6.7, 3.6, 5.9],
            [4.0, 3.6, 2.9],
            [5.3, 3.3, 6.1],
            [5.7, 5.5, 6.2],
            [5.2, 4.1, 6.4],
            [2.4, 2.9, 3.2],
            [3.2, 2.7, 4.0],
            [2.7, 2.6, 4.1],
        ]
    )
    M = MVNEntropy(x)
    # print(M)
    # print("%.15g %.15g %.15g"%(M.p_kurtosis, M.p_skewness, M.entropy))
    assert abs(M.p_kurtosis - 0.265317890462476) <= 1e-10
    assert abs(M.p_skewness - 0.773508066109368) <= 1e-10
    assert abs(M.entropy - 5.7920040570988) <= 1e-10

    ## wnn_entropy doesn't work for small sample sizes (no surprise there!)
    # S_wnn, Serr_wnn = wnn_entropy(x)
    # assert abs(S_wnn - 5.7920040570988) <= 1e-10
    # print("wnn %.15g, target %g"%(S_wnn, 5.7920040570988))


def demo():
    # hide module load time from Timer
    from sklearn.neighbors import NearestNeighbors

    ## Bootstrap didn't help, but leave the test code in place for now
    # D = Dirichlet(alpha=[0.02]*20)
    # theta = D.rvs(size=1000)
    # S, Serr = wnn_bootstrap(D.rvs(size=200000))
    # print("bootstrap", S, D.entropy())
    # return
    if False:
        # Multivariate T distribution
        D = stats.t(df=4)
        _show_entropy("T;df=4", D, N=20000)
        D = MultivariateT(sigma=np.diag([1]), df=4)
        _show_entropy("MT[1];df=4", D, N=20000)
        D = MultivariateT(sigma=np.diag([1, 12, 0.2]) ** 2, df=4)
        _show_entropy("MT[1,12,0.2];df=4", D, N=10000)
        D = MultivariateT(sigma=np.diag([1] * 10), df=4)
        _show_entropy("MT[1]*10;df=4", D, N=10000)
        D = MultivariateT(sigma=np.diag([1, 12, 0.2, 1e2, 1e-2, 1]) ** 2, df=4)
        _show_entropy("MT[1,12,0.2,1e3,1e-3,1];df=4", D, N=10000)
        return

    if False:
        # Multivariate skew normal distribution
        D = stats.skewnorm(5)
        _show_entropy("skew=5 N[1]", D, N=20000)
        D = Joint(stats.skewnorm(5, 0, s) for s in [1, 12, 0.2])
        _show_entropy("skew=5 N[1,12,0.2]", D, N=10000)
        D = Joint(stats.skewnorm(5, 0, s) for s in [1] * 10)
        _show_entropy("skew=5 N[1]*10", D, N=10000)
        D = Joint(stats.skewnorm(5, 0, s) for s in [1, 12, 0.2, 1e2, 1e-2, 1])
        _show_entropy("skew=5 N[1,12,0.2,1e3,1e-3,1]", D, N=10000)
        # print("double check entropy", D.entropy()/LN2, entropy_mc(D)/LN2)
        return

    D = Box(center=[100] * 10, width=np.linspace(1, 10, 10))
    _show_entropy("Box 10!", D, N=10000)
    D = stats.norm(10, 8)
    # _show_entropy("N[100,8]", D, N=100)
    # _show_entropy("N[100,8]", D, N=200)
    # _show_entropy("N[100,8]", D, N=500)
    # _show_entropy("N[100,8]", D, N=1000)
    # _show_entropy("N[100,8]", D, N=2000)
    # _show_entropy("N[100,8]", D, N=5000)
    _show_entropy("N[100,8]", D, N=10000)
    # _show_entropy("N[100,8]", D, N=20000)
    # _show_entropy("N[100,8]", D, N=50000)
    # _show_entropy("N[100,8]", D, N=100000)
    D = stats.multivariate_normal(cov=np.diag([1, 12, 0.2]) ** 2)
    # _show_entropy("MVN[1,12,0.2]", D)
    D = stats.multivariate_normal(cov=np.diag([1] * 10) ** 2)
    # _show_entropy("MVN[1]*10", D, N=1000)
    _show_entropy("MVN[1]*10", D, N=10000)
    # _show_entropy("MVN[1]*10", D, N=100000)
    # _show_entropy("MVN[1]*10", D, N=200000, N_entropy=20000)
    D = stats.multivariate_normal(cov=np.diag([1, 12, 0.2, 1, 1, 1]) ** 2)
    # _show_entropy("MVN[1,12,0.2,1,1,1]", D, N=100)
    # _show_entropy("MVN[1,12,0.2,1,1,1]", D, N=1000)
    _show_entropy("MVN[1,12,0.2,1,1,1]", D, N=10000)
    # _show_entropy("MVN[1,12,0.2,1,1,1]", D, N=100000)
    D = stats.multivariate_normal(cov=np.diag([1, 12, 0.2, 1e2, 1e-2, 1]) ** 2)
    # _show_entropy("MVN[1,12,0.2,1e3,1e-3,1]", D, N=100)
    # _show_entropy("MVN[1,12,0.2,1e3,1e-3,1]", D, N=1000)
    _show_entropy("MVN[1,12,0.2,1e3,1e-3,1]", D, N=10000)
    # _show_entropy("MVN[1,12,0.2,1e3,1e-3,1]", D, N=100000)
    D = GaussianMixture([1, 10], mu=[[0] * 10, [100] * 10], sigma=[[10] * 10, [0.1] * 10])
    _show_entropy("bimodal mixture", D)
    D = Dirichlet(alpha=[0.02] * 20)
    # _show_entropy("Dirichlet[0.02]*20", D, N=1000)
    # _show_entropy("Dirichlet[0.02]*20", D, N=2000)
    # _show_entropy("Dirichlet[0.02]*20", D, N=5000)
    # _show_entropy("Dirichlet[0.02]*20", D, N=10000)
    _show_entropy("Dirichlet[0.02]*20", D, N=20000)
    # _show_entropy("Dirichlet[0.02]*20", D, N=50000)
    # _show_entropy("Dirichlet[0.02]*20", D, N=200000, N_entropy=20000)


if __name__ == "__main__":  # pragma: no cover
    demo()



import wx

IS_MAC = wx.Platform == "__WXMAC__"


class LogView(wx.Panel):
    title = "Log"
    default_size = (600, 200)

    def __init__(self, *args, **kw):
        wx.Panel.__init__(self, *args, **kw)

        self.log_info = []
        vsizer = wx.BoxSizer(wx.VERTICAL)

        self.progress = wx.TextCtrl(self, -1, style=wx.TE_MULTILINE | wx.HSCROLL)
        self._redraw()

        vsizer.Add(self.progress, 1, wx.EXPAND)

        self.SetSizer(vsizer)
        vsizer.Fit(self)
        self.SetAutoLayout(True)
        # self.SetupScrolling()

        self.Bind(wx.EVT_SHOW, self.OnShow)

    def OnShow(self, event):
        if not event.Show:
            return
        # print "showing log"
        if self._need_redraw:
            # print "-redraw"
            self._redraw()

    def get_state(self):
        return self.log_info

    def set_state(self, state):
        self.log_info = state
        self._redraw()

    def log_message(self, message):
        if len(self.log_info) > 1000:
            del self.log_info[:-1000]
        self.log_info.append(message)
        self._redraw()

    def _redraw(self):
        if not IS_MAC and not self.IsShown():
            self._need_redraw = True
        else:
            self._need_redraw = False
            self.progress.Clear()
            self.progress.AppendText("\n".join(self.log_info))



"""
Monotonic spline modeling.
"""

__all__ = ["monospline", "hermite", "count_inflections", "plot_inflections"]

import numpy as np
from numpy import diff, hstack, sqrt, searchsorted, asarray, nonzero, linspace, isnan


def monospline(x, y, xt):
    r"""
    Monotonic cubic hermite interpolation.

    Returns $p(x_t)$ where $p(x_i)= y_i$ and $p(x) \leq p(x_i)$
    if $y_i \leq y_{i+1}$ for all $y_i$.  Also works for decreasing
    values $y$, resulting in decreasing $p(x)$.  If $y$ is not monotonic,
    then $p(x)$ may peak higher than any $y$, so this function is not
    suitable for a strict constraint on the interpolated function when
    $y$ values are unconstrained.

    http://en.wikipedia.org/wiki/Monotone_cubic_interpolation
    """
    with np.errstate(all="ignore"):
        x = hstack((x[0] - 1, x, x[-1] + 1))
        y = hstack((y[0], y, y[-1]))
        dx = diff(x)
        dy = diff(y)
        dx[abs(dx) < 1e-10] = 1e-10
        delta = dy / dx
        m = (delta[1:] + delta[:-1]) / 2
        m = hstack((0, m, 0))
        alpha, beta = m[:-1] / delta, m[1:] / delta
        d = alpha**2 + beta**2

        # print "ma",m
        for i in range(len(m) - 1):
            if isnan(delta[i]):
                m[i] = delta[i + 1]
            elif dy[i] == 0 or alpha[i] == 0 or beta[i] == 0:
                m[i] = m[i + 1] = 0
            elif d[i] > 9:
                tau = 3.0 / sqrt(d[i])
                m[i] = tau * alpha[i] * delta[i]
                m[i + 1] = tau * beta[i] * delta[i]
                # if isnan(m[i]) or isnan(m[i+1]):
                #    print i,"isnan",tau,d[i], alpha[i],beta[i],delta[i]
            # elif isnan(m[i]):
            #    print i,"isnan",delta[i],dy[i]
            # m[ dy[1:]*dy[:-1]<0 ] = 0
        # if np.any(isnan(m)|isinf(m)):
        #    print "mono still has bad values"
        #    print "m",m
        #    print "delta",delta
        #    print "dx,dy",list(zip(dx,dy))
        #    m[isnan(m)|isinf(m)] = 0

    return hermite(x, y, m, xt)


def hermite(x, y, m, xt):
    """
    Computes the cubic hermite polynomial $p(x_t)$.

    The polynomial goes through all points $(x_i,y_i)$ with slope
    $m_i$ at the point.
    """
    with np.errstate(all="ignore"):
        x, y, m, xt = [asarray(v, "d") for v in (x, y, m, xt)]
        idx = searchsorted(x[1:-1], xt)
        h = x[idx + 1] - x[idx]
        h[h <= 1e-10] = 1e-10
        s = (y[idx + 1] - y[idx]) / h
        v = xt - x[idx]
        c3, c2, c1, c0 = ((m[idx] + m[idx + 1] - 2 * s) / h**2, (3 * s - 2 * m[idx] - m[idx + 1]) / h, m[idx], y[idx])
    return ((c3 * v + c2) * v + c1) * v + c0


# TODO: move inflection point code to data.py
def count_inflections(x, y):
    """
    Count the number of inflection points in a curve.
    """
    with np.errstate(all="ignore"):
        m = (y[2:] - y[:-2]) / (x[2:] - x[:-2])
        b = y[2:] - m * x[2:]
        delta = y[1:-1] - (m * x[1:-1] + b)
        delta = delta[nonzero(delta)]  # ignore points on the line

    sign_change = (delta[1:] * delta[:-1]) < 0
    return sum(sign_change)


def plot_inflections(x, y):
    """
    Plot inflection points in a curve.
    """
    m = (y[2:] - y[:-2]) / (x[2:] - x[:-2])
    b = y[2:] - m * x[2:]
    delta = y[1:-1] - (m * x[1:-1] + b)
    t = linspace(x[0], x[-1], 400)
    import pylab

    ax1 = pylab.subplot(211)
    pylab.plot(t, monospline(x, y, t), "-b", x, y, "ob")
    pylab.subplot(212, sharex=ax1)
    delta_x = x[1:-1]
    pylab.stem(delta_x, delta)
    pylab.plot(delta_x[delta < 0], delta[delta < 0], "og")
    pylab.axis([x[0], x[-1], min(min(delta), 0), max(max(delta), 0)])



from copy import deepcopy
from threading import Thread
from threading import Event
import traceback

from blinker import Signal

import numpy as np
from bumps import monitor
from bumps.fitters import FitDriver, nllf_scale, format_uncertainty
from bumps.mapper import MPMapper, SerialMapper, can_pickle
from bumps.util import redirect_console
from bumps.history import History

# from .convergence_view import ConvergenceMonitor
# ==============================================================================

PROGRESS_DELAY = 5
IMPROVEMENT_DELAY = 5

EVT_FIT_PROGRESS = Signal()
EVT_FIT_COMPLETE = Signal()

# NOTE: GUI monitors are running in a separate thread.  They should not
# touch the problem internals.


class GUIProgressMonitor(monitor.TimedUpdate):
    def __init__(self, problem, progress=None, improvement=None):
        monitor.TimedUpdate.__init__(
            self, progress=progress or PROGRESS_DELAY, improvement=improvement or IMPROVEMENT_DELAY
        )
        self.problem = problem

    def show_progress(self, history):
        scale, err = nllf_scale(self.problem)
        chisq = format_uncertainty(scale * history.value[0], err)
        evt = dict(
            # problem=self.problem,
            message="progress",
            step=history.step[0],
            value=history.value[0],
            chisq=chisq,
            point=history.point[0] + 0,
        )  # avoid race
        EVT_FIT_PROGRESS.send(evt)

    def show_improvement(self, history):
        evt = dict(
            message="improvement", step=history.step[0], value=history.value[0], point=history.point[0] + 0
        )  # avoid race
        EVT_FIT_PROGRESS.send(evt)


class ConvergenceMonitor(monitor.Monitor):
    """
    Generic GUI monitor for fitting.

    Sends a convergence_update event every *rate*
    seconds.  Gathers statistics about the best, worst, median and +/- 1 interquartile
    range.  This will be the input for the convergence plot.

    *problem* should be the fit problem handed to the fit thread, and not
    a copy. This is because it is used for direct comparison with the current
    fit object in the progress panels so that stray messages don't cause
    confusion when making graphs.

    *message* is a dispatch string used by the OnFitProgress event processor
    in the app to determine which progress panel should receive the event.
    """

    def __init__(self, problem, message="convergence_update", rate=0):
        self.time = 0
        self.rate = rate  # rate=0 for no progress update, only final
        self.problem = problem
        self.message = message
        self.pop = []

    def config_history(self, history):
        history.requires(population_values=1, value=1)
        history.requires(time=1)
        # history.requires(time=1, population_values=1, value=1)

    def __call__(self, history):
        # from old ConvergenceMonitor:
        best = history.value[0]
        try:
            pop = history.population_values[0]
            n = len(pop)
            p = np.sort(pop)
            (
                QI,
                Qmid,
            ) = int(0.2 * n), int(0.5 * n)
            self.pop.append((best, p[0], p[QI], p[Qmid], p[-1 - QI], p[-1]))
        except (AttributeError, TypeError):
            self.pop.append((best,))

        if self.rate > 0 and history.time[0] >= self.time + self.rate:
            evt = dict(message=self.message, pop=self.progress())
            EVT_FIT_PROGRESS.send(evt)
            self.time = history.time[0]

    def progress(self):
        return np.empty((0, 1), "d") if not self.pop else np.array(self.pop)

    def final(self):
        """
        Close out the monitor
        """
        evt = dict(message=self.message, pop=self.progress())
        EVT_FIT_PROGRESS.send(evt)


# Horrible hacks:
# (1) We are grabbing uncertainty_state from the history on monitor update
# and holding onto it for the monitor final call. Need to restructure the
# history/monitor interaction so that object lifetimes are better controlled.


class DreamMonitor(monitor.Monitor):
    def __init__(self, problem, message, fitter, rate=0):
        self.time = 0
        self.rate = rate  # rate=0 for no progress update, only final
        self.update_counter = 0
        self.problem = problem
        self.fitter = fitter
        self.message = message
        self.uncertainty_state = None
        # emit None uncertainty state to start with
        evt = dict(
            message=message,
            uncertainty_state=None,
        )
        EVT_FIT_PROGRESS.send(evt)

    def config_history(self, history):
        history.requires(time=1)

    def __call__(self, history):
        self.uncertainty_state = getattr(history, "uncertainty_state", None)
        self.time = history.time[0]
        if self.rate > 0:
            update_counter = history.time[0] // self.rate
            if update_counter > self.update_counter:
                self.update_counter = update_counter
                evt = dict(
                    message=self.message,
                    time=self.time,
                    uncertainty_state=deepcopy(self.uncertainty_state),
                )
                EVT_FIT_PROGRESS.send(evt)

    def final(self):
        """
        Close out the monitor
        """
        # Note: win.uncertainty_state protected by win.fit_lock
        # self.win.uncertainty_state = self.uncertainty_state
        evt = dict(
            message="uncertainty_final",
            time=self.time,
            uncertainty_state=deepcopy(self.uncertainty_state),
        )
        EVT_FIT_PROGRESS.send(evt)


# ==============================================================================


class FitThread(Thread):
    """Run the fit in a separate thread from the GUI thread."""

    def __init__(
        self,
        abort_event: Event,
        problem=None,
        fitclass=None,
        options=None,
        mapper=None,
        parallel=0,
        convergence_update=5,
        uncertainty_update=300,
        terminate_on_finish=False,
    ):
        # base class initialization
        # Process.__init__(self)

        Thread.__init__(self)
        self.abort_event = abort_event
        self.problem = problem
        self.fitclass = fitclass
        self.options = options if isinstance(options, dict) else dict()
        self.mapper = mapper
        self.parallel = parallel
        self.convergence_update = convergence_update
        self.uncertainty_update = uncertainty_update
        self.terminate_on_finish = terminate_on_finish

    def abort_test(self):
        return self.abort_event.is_set()

    def run(self):
        # TODO: we have no interlocks on changes in problem state.  What
        # happens when the user changes the problem while a fit is being run?
        # May want to keep a history of changes to the problem definition,
        # along with a function to reverse them so we can handle undo.

        # NOTE: Problem must be the original problem (not a copy) when used
        # inside the GUI monitor otherwise AppPanel will not be able to
        # recognize that it is the same problem when updating views.
        try:
            monitors = [
                GUIProgressMonitor(self.problem),
                ConvergenceMonitor(self.problem, rate=self.convergence_update),
                # GUIMonitor(self.problem,
                #            message="convergence_update",
                #            monitor=ConvergenceMonitor(),
                #            rate=self.convergence_update),
                DreamMonitor(
                    self.problem, fitter=self.fitclass, message="uncertainty_update", rate=self.uncertainty_update
                ),
            ]

            mapper = self.mapper
            if mapper is None:
                # fallback options if no mapper is specified
                mp_available = True
                try:
                    import _multiprocessing
                except ImportError:
                    mp_available = False
                # Only use MPMapper if the problem can be pickled
                # and multiprocessing is available, and parallel != 1
                mapper = MPMapper if mp_available and self.parallel != 1 and can_pickle(self.problem) else SerialMapper

            # Be safe and send a private copy of the problem to the fitting engine
            # print "fitclass",self.fitclass
            problem = deepcopy(self.problem)
            # print "fitclass id",id(self.fitclass),self.fitclass,threading.current_thread()
            driver = FitDriver(
                self.fitclass,
                problem=problem,
                monitors=monitors,
                abort_test=self.abort_test,
                mapper=mapper.start_mapper(problem, [], cpus=self.parallel),
                **self.options,
            )

            x, fx = driver.fit()
            # Give final state message from monitors
            for M in monitors:
                if hasattr(M, "final"):
                    M.final()

            with redirect_console() as fid:
                driver.show()
                captured_output = fid.getvalue()

            evt = dict(
                message="complete",
                problem=self.problem,
                point=x,
                value=fx,
                info=captured_output,
                fitter_id=self.fitclass.id,
            )
            EVT_FIT_COMPLETE.send(evt)
            self.result = evt

        except Exception as exc:
            tb = "".join(traceback.TracebackException.from_exception(exc).format())
            evt = dict(message="error", error_string=str(exc), traceback=tb)
            EVT_FIT_COMPLETE.send(evt)



# Copyright (C) 2006-2011, University of Maryland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Author: Nikunj Patel

"""
This module implements the Summary View panel.
"""

# ==============================================================================

import wx
import wx.lib.scrolledpanel as scrolled

from . import signal
from .util import nice

IS_MAC = wx.Platform == "__WXMAC__"

NUMPIX = 400
NUMTICKS = NUMPIX * 5 - 1
COMPACTIFY_VERTICAL = {
    "__WXMAC__": 2,
    "__WXGTK__": 6,
    "__WXMSW__": 4,
}.get(wx.Platform, 6)


class SummaryView(scrolled.ScrolledPanel):
    """
    Model view showing summary of fit (only fittable parameters).
    """

    title = "Summary"
    default_size = (600, 500)

    def __init__(self, *args, **kw):
        scrolled.ScrolledPanel.__init__(self, *args, **kw)

        self.display_list = []
        self.parameters = []

        self.sizer = wx.GridBagSizer(hgap=0, vgap=-COMPACTIFY_VERTICAL)
        self.SetSizer(self.sizer)
        self.sizer.Fit(self)

        self.SetAutoLayout(True)
        self.SetupScrolling()

        self._need_update_parameters = self._need_update_model = False
        self.Bind(wx.EVT_SHOW, self.OnShow)

    def OnShow(self, event):
        # print "show event"
        if not event.Show:
            return
        # print "showing summary"
        if self._need_update_model:
            # print "-update_model"
            self.update_model(self.model)
        elif self._need_update_parameters:
            # print "-update_parameters"
            self.update_parameters(self.model)
        event.Skip()

    # ============ Operations on the model  ===============

    def get_state(self):
        return self.model

    def set_state(self, state):
        self.set_model(state)

    def set_model(self, model):
        self.model = model
        self.update_model(model)

    def update_model(self, model):
        if self.model != model:
            return

        if not IS_MAC and not self.IsShown():
            # print "summary not shown config"
            self._need_update_model = True
        else:
            # print "summary shown config"
            self._update_model()
            self._need_update_parameters = self._need_update_model = False

    def update_parameters(self, model):
        if self.model != model:
            return
        if not IS_MAC and not self.IsShown():
            # print "summary not shown update"
            self._need_update_parameters = True
        else:
            # print "summary shown upate"
            self._need_update_parameters = False
            self._update_parameters()

    def _update_model(self):
        # print "drawing"
        self.parameters = self.model._parameters if self.model is not None else []
        self.sizer.Clear(True)
        # self.sizer.Clear()
        self.display_list = []

        self.layer_label = wx.StaticText(self, wx.ID_ANY, "Fit Parameter", size=(160, -1))
        self.slider_label = wx.StaticText(self, wx.ID_ANY, "", size=(NUMPIX, -1))
        self.value_label = wx.StaticText(self, wx.ID_ANY, "Value", size=(100, -1))
        self.low_label = wx.StaticText(self, wx.ID_ANY, "Minimum", size=(100, -1))
        self.high_label = wx.StaticText(self, wx.ID_ANY, "Maximum", size=(100, -1))

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        hbox.Add(self.layer_label, 0, wx.LEFT, 1)
        hbox.Add(self.slider_label, 0, wx.LEFT, 1)
        hbox.Add(self.value_label, 0, wx.LEFT, 21)
        hbox.Add(self.low_label, 0, wx.LEFT, 1)
        hbox.Add(self.high_label, 0, wx.LEFT, 1)

        # Note that row at pos=(0,0) is not used to add a blank row.
        self.sizer.Add(hbox, pos=(1, 0), flag=wx.BOTTOM, border=COMPACTIFY_VERTICAL)

        line = wx.StaticLine(self, wx.ID_ANY)
        self.sizer.Add(line, pos=(2, 0), flag=wx.EXPAND | wx.RIGHT | wx.BOTTOM, border=COMPACTIFY_VERTICAL)

        # TODO: better interface to fittable parameters
        if self.model is not None:
            pars = self.model._parameters
            # pars = sorted(pars, cmp=lambda x,y: cmp(x.name, y.name))
            for p in pars:
                self.display_list.append(ParameterSummary(self, p, self.model))

        for index, item in enumerate(self.display_list):
            self.sizer.Add(item, pos=(index + 3, 0))

        self.SetupScrolling()
        self.Layout()

    def _update_parameters(self):
        # print "updating"
        if self.parameters is not self.model._parameters:
            self._update_model()
        else:
            for p in self.display_list:
                p.update_slider()


VALUE_PRECISION = 6
VALUE_FORMAT = "{{:.{:d}g}}".format(VALUE_PRECISION)


class ParameterSummary(wx.Panel):
    """Build one parameter line for display."""

    def __init__(self, parent, parameter, model):
        wx.Panel.__init__(self, parent=parent, id=wx.ID_ANY)

        self.parameter = parameter
        self.model = model

        self.low, self.high = (v for v in self.parameter.prior.limits)

        text_hbox = wx.BoxSizer(wx.HORIZONTAL)

        self.layer_name = wx.StaticText(self, wx.ID_ANY, str(self.parameter.name), size=(160, -1), style=wx.TE_LEFT)
        self.slider = wx.Slider(
            self, wx.ID_ANY, value=0, minValue=0, maxValue=NUMPIX * 5 - 1, size=(NUMPIX, -1), style=wx.SL_HORIZONTAL
        )
        self.value = wx.StaticText(
            self, wx.ID_ANY, VALUE_FORMAT.format(self.parameter.value), size=(100, -1), style=wx.TE_LEFT
        )
        self.min_range = wx.StaticText(self, wx.ID_ANY, VALUE_FORMAT.format(self.low), size=(100, -1), style=wx.TE_LEFT)
        self.max_range = wx.StaticText(
            self, wx.ID_ANY, VALUE_FORMAT.format(self.high), size=(100, -1), style=wx.TE_LEFT
        )

        # Add text strings and slider to sizer.
        text_hbox_flags = wx.LEFT | wx.ALIGN_CENTER_VERTICAL
        text_hbox.Add(self.layer_name, 0, text_hbox_flags, 1)
        text_hbox.Add(self.slider, 0, text_hbox_flags, 1)
        text_hbox.Add(self.value, 0, text_hbox_flags, 21)
        text_hbox.Add(self.min_range, 0, text_hbox_flags, 1)
        text_hbox.Add(self.max_range, 0, text_hbox_flags, 1)

        self.SetSizer(text_hbox)

        self.slider.Bind(wx.EVT_SCROLL, self.OnScroll)
        self.update_slider()

    def update_slider(self):
        slider_pos = int(self.parameter.prior.get01(self.parameter.value) * NUMTICKS)
        # Add line below if get01 doesn't protect against values out of range.
        # slider_pos = min(max(slider_pos,0),100)
        self.slider.SetValue(slider_pos)
        self.value.SetLabel(VALUE_FORMAT.format(nice(self.parameter.value, digits=VALUE_PRECISION)))

        # Update new min and max range of values if changed.
        newlow, newhigh = (v for v in self.parameter.prior.limits)
        if newlow != self.low:
            self.min_range.SetLabel(VALUE_FORMAT.format(newlow))

        # if newhigh != self.high:
        self.max_range.SetLabel(VALUE_FORMAT.format(newhigh))

    def OnScroll(self, event):
        value = self.slider.GetValue()
        new_value = self.parameter.prior.put01(value / NUMTICKS)
        nice_new_value = nice(new_value, digits=VALUE_PRECISION)
        self.parameter.value = nice_new_value
        self.value.SetLabel(VALUE_FORMAT.format(nice_new_value))
        signal.update_parameters(model=self.model, delay=1)



"""
MCMC step acceptance test.
"""

__all__ = ["metropolis", "metropolis_dr"]

import os

from numpy import array, cov, dot, errstate, exp, eye, minimum, sqrt, where
from numpy.linalg import cholesky, inv, norm

from . import util

BUMPS_TEMPERATURE = float(os.environ.get("BUMPS_TEMPERATURE", "1"))


def paccept(logp_old, logp_try):
    """
    Returns the probability of taking a metropolis step given two
    log density values.
    """
    return exp(minimum(logp_try - logp_old, 0) / BUMPS_TEMPERATURE)


def metropolis(xtry, logp_try, xold, logp_old, step_alpha):
    """
    Metropolis rule for acceptance or rejection

    Generates the next generation, *newgen* from::

        x_new[k] = x[k]     if U > alpha
                 = x_old[k] if U <= alpha

    where alpha is p/p_old and accept is U > alpha.

    Returns x_new, logp_new, alpha, accept
    """
    with errstate(under="ignore"):
        alpha = paccept(logp_try=logp_try, logp_old=logp_old)
        alpha *= step_alpha
    accept = alpha > util.rng.rand(*alpha.shape)
    logp_new = where(accept, logp_try, logp_old)
    ## The following only works for vectors:
    # xnew = where(accept, xtry, xold)
    xnew = xtry + 0
    for i, a in enumerate(accept):
        if not a:
            xnew[i] = xold[i]

    return xnew, logp_new, alpha, accept


def dr_step(x, scale):
    """
    Delayed rejection step.
    """

    # Compute the Cholesky Decomposition of X
    nchains, npars = x.shape
    r = (2.38 / sqrt(npars)) * cholesky(cov(x.T) + 1e-5 * eye(npars))

    # Now do a delayed rejection step for each chain
    delta_x = dot(util.rng.randn(*x.shape), r) / scale

    # Generate ergodicity term
    eps = 1e-6 * util.rng.randn(*x.shape)

    # Update x_old with delta_x and eps;
    return x + delta_x + eps, r


def metropolis_dr(xtry, logp_try, x, logp, xold, logp_old, alpha12, R):
    """
    Delayed rejection metropolis
    """

    # Compute alpha32 (note we turned x and xtry around!)
    alpha32 = paccept(logp_try=logp, logp_old=logp_try)

    # Calculate alpha for each chain
    l2 = paccept(logp_try=logp_try, logp_old=logp_old)
    iR = inv(R)
    q1 = array(
        [exp(-0.5 * (norm(dot(x2 - x1, iR)) ** 2 - norm(dot(x1 - x0, iR)) ** 2)) for x0, x1, x2 in zip(xold, x, xtry)]
    )
    alpha13 = l2 * q1 * (1 - alpha32) / (1 - alpha12)

    accept = alpha13 > util.rng.rand(*alpha13.shape)
    logp_new = where(accept, logp_try, logp)
    ## The following only works for vectors:
    # xnew = where(accept, xtry, x)
    xnew = xtry + 0
    for i, a in enumerate(accept):
        if not a:
            xnew[i] = x[i]

    return xnew, logp_new, alpha13, accept



"""
Option parser for bumps command line
"""

import sys

import numpy as np

from .fitters import FITTERS, FIT_AVAILABLE_IDS, FIT_ACTIVE_IDS, FIT_DEFAULT_ID


# TODO: replace with standard argparse module
class ParseOpts(object):
    """
    Options parser.

    Subclass should define *MINARGS*, *FLAGS*, *VALUES* and *USAGE*.

    *MINARGS* is the minimum number of positional arguments.

    *FLAGS* is a set of arguments that may be present or absent.

    *VALUES* is a set of arguments that take values.  Value checking
    can be done in the setter for each argument in the set.  Default
    values should be set in the corresponding object attribute.

    *USAGE* is the help string to display for option "help".

    The constructor will invoke the command line parser, leaving the
    values set by the command line as attribute values.   Flag options
    will be True or False.
    """

    MINARGS = 0
    FLAGS = set()
    VALUES = set()
    #: Value to use if a value flag is is present without '='.  This is
    #: different from the default value if the flag is not present, which
    #: is the default value set in the calling class.
    IMPLICIT_VALUES = {}
    USAGE = ""

    def __init__(self, args):
        if self.VALUES & self.FLAGS:
            raise TypeError("option used as both a flag and a value: %s" % ",".join(self.VALUES & self.FLAGS))
        self._parse(args)

    def _parse(self, args):
        if any(v in args for v in ("-?", "-h", "-help")):
            print(self.USAGE)
            sys.exit()

        # Drop the "bumps" arg from the beginning of the list.
        args = args[1:]

        # Fill in implicit values. We need to do this to support something
        # like "bumps ... --parallel=2 ... --parallel", which should have the
        # later (implicit) parameter take precedence over the earlier
        # parameters.
        args = [
            (arg + "=" + str(self.IMPLICIT_VALUES[arg[2:]]) if arg[2:] in self.IMPLICIT_VALUES else arg) for arg in args
        ]

        # Parse options.
        # Given tuples [..., (a, 1), ..., (a, 2), ...], then dict(tuples)
        # will use the later value for the key rather than the earlier
        # value, which is what we want for the command line interpreter.
        position_args = [v for v in sys.argv[1:] if not v.startswith("--")]
        flag_args = [
            v[2:]  # convert --flag => flag
            for v in args
            if v.startswith("--") and not "=" in v
        ]
        value_args = dict(
            v[2:].split("=", 1)  # convert --flag=value => (flag, value)
            for v in args
            if v.startswith("--") and "=" in v
        )

        # Check that options are valid.
        # TODO: move type checking from FitConfig.set_from_cli to here.
        flags, values = set(flag_args), set(value_args.keys())
        unknown = (flags | values) - (self.FLAGS | self.VALUES)
        unexpected_value = flags - self.FLAGS
        blank_values = set(k for k, v in value_args.items() if k in self.VALUES and v == "")
        missing_value = (values - self.VALUES) | blank_values
        errors = []
        if any(unknown):
            errors.append("Unknown options --%s." % ", --".join(unknown))
        if any(unexpected_value):
            errors.append("Unexpected value for --%s." % ", --".join(sorted(unexpected_value)))
        if any(missing_value):
            errors.append("Missing value for --%s." % ", --".join(sorted(missing_value)))
        if errors:
            message = " ".join(errors + ["Use -? for help."])
            raise ValueError(message)

        # Set the values into the fields.
        for option in self.FLAGS:
            setattr(self, option, (option in flags))

        for option, value in value_args.items():
            setattr(self, option, value)

        self.args = position_args


# === Fitter option parsing ===


class ChoiceList(object):
    def __init__(self, *choices):
        self.choices = choices

    def __call__(self, value):
        if not value in self.choices:
            raise ValueError('invalid option "%s": use %s' % (value, "|".join(self.choices)))
        else:
            return value


def yesno(value):
    if value.lower() in ("true", "yes", "on", "1"):
        return True
    elif value.lower() in ("false", "no", "off", "0"):
        return False
    raise ValueError('invalid option "%s": use yes|no')


def parse_int(value):
    float_value = float(value)
    if int(float_value) != float_value:
        raise ValueError("integer expected")
    return int(float_value)


FIT_FIELDS = dict(
    starts=("Starts", parse_int),
    steps=("Steps", parse_int),
    samples=("Samples", parse_int),
    xtol=("x tolerance", float),
    ftol=("f(x) tolerance", float),
    alpha=("Convergence", float),
    stop=("Stopping criteria", str),
    thin=("Thinning", parse_int),
    burn=("Burn-in steps", parse_int),
    pop=("Population", float),
    init=("Initializer", ChoiceList("eps", "lhs", "cov", "random")),
    CR=("Crossover ratio", float),
    F=("Scale", float),
    nT=("# Temperatures", parse_int),
    Tmin=("Min temperature", float),
    Tmax=("Max temperature", float),
    radius=("Simplex radius", float),
    # TODO: convert --trim into a boolean flag and update docs
    trim=("Burn-in trim", yesno),
    outliers=("Outliers", ChoiceList("none", "iqr", "grubbs", "mahal")),
)

# Make sure all settings are parseable
for fit in FITTERS:
    assert all(opt in FIT_FIELDS for opt, _ in fit.settings), "Fitter %s contains unknown settings %s" % (
        fit.id,
        ", ".join(opt for opt, _ in sorted(fit.settings) if opt not in FIT_FIELDS),
    )
del fit


class FitConfig(object):
    """
    Fit settings configuration object.

    The command line parser will define a FitConfig object which contains
    the fitter that was given on the command line and all its options.  For
    embedded bumps, which does not use the bumps command line parser, a
    new FitConfig object can be created with its own selected options.

    **Attributes**

    *ids = [id, id, ...]* is a list available fitters in "preferred" order.
    Depending on usage, you may want to sort them, or alternatively, sort
    by long name with *[id for _,id in sorted((v,k) for k,v in self.names]*

    *fitters = {id: fitclasss}* maps ids to fitters.

    *names* = {id: name}* maps ids to long names

    *settings = {id: [(option, default), ...]}* maps ids to default settings.
    The order of the settings is the preferred order to present the settings
    to the user in a GUI dialog for example.

    *values = {id: {option: value, ...}}* maps ids to the settings for
    each fitter.  Note that in the GUI, different fitters may have there
    settings recorded and preserved even when not selected.

    *active_ids = [id, id, ...]* is the list of fitters to show the user in
    a GUI dialog for example.  The other fitters should still be available from
    the command line.

    *default_id = id* is the fitter to use by default.

    *selected_id = id* is the fitter that was selected, either by command line
    or by GUI.

    *selected_values = {option: value}* returns the settings for the current
    fitter.

    *selected_name = name* returns the name of the selected fitter.

    *selected_fitter = FitClass* returns the class of the selected fitter.

    """

    def __init__(self, default=FIT_DEFAULT_ID, active=FIT_ACTIVE_IDS):
        # Keep a private copy of the configure settings rather than modifying
        # the global defaults
        self.ids = [fit.id for fit in FITTERS]
        # FITTERS is a list of FitBase classes
        # Each class has:
        #     fit.id: the short name used on the command line
        #     fit.name: the long name used in the GUI
        #     fit.settings: available options: [(key,default value), ...]
        self.fitters = dict((fit.id, fit) for fit in FITTERS)
        self.names = dict((fit.id, fit.name) for fit in FITTERS)
        self.settings = dict((fit.id, fit.settings) for fit in FITTERS)
        self.values = dict((fit.id, dict(fit.settings)) for fit in FITTERS)
        if not all(k in self.ids for k in active):
            raise ValueError("Some active fitters are not available")
        if default not in active:
            raise ValueError("default fitter is not active")
        self.active_ids = active
        self.default_id = default
        self.selected_id = default

    def set_from_cli(self, opts):
        """
        Use the BumpsOpts command line parser values to set the selected
        fitter and its configuration options.
        """
        fitter = opts.fit
        self.selected_id = fitter
        # Convert supplied options to the correct types and save them in value
        for field, reset_value in self.settings[fitter]:
            value = getattr(opts, field, None)
            parse = FIT_FIELDS[field][1]
            if value is not None:
                try:
                    self.values[fitter][field] = parse(value)
                except Exception as exc:
                    raise ValueError("error in --%s: %s" % (field, str(exc)))
                    # print("options=%s"%(str(self.options)))

    @property
    def selected_values(self):
        return self.values[self.selected_id]

    @property
    def selected_name(self):
        return self.names[self.selected_id]

    @property
    def selected_fitter(self):
        return self.fitters[self.selected_id]


#: FitConfig singleton for the common case in which only one config is needed.
#: There may be other use cases, such as saving the fit config along with the
#: rest of the state so that on resume the fit options are restored, but in that
#: case the application will not be using the singleton.
FIT_CONFIG = FitConfig()


# === Bumps options parsing ===
class BumpsOpts(ParseOpts):
    """
    Option parser for bumps.
    """

    MINARGS = 1
    # TODO: document all options in USAGE and doc/guide/options.rst
    # TODO: remove application-specific options like --staj
    FLAGS = set(
        (
            "preview",
            "chisq",
            "profile",
            "time_model",
            "simulate",
            "simrandom",
            "shake",
            "worker",  # internal, so not documented
            "multiprocessing-fork",  # passed in when app is a frozen image
            "remote",  # not active, so not documented
            "batch",
            "noshow",
            "overwrite",
            "stepmon",
            "err",
            "cov",
            "edit",
            "mpi",
            "keep_best",
            "staj",
            # passed when not running bumps, but instead using a
            # bundled application as a python distribution with domain
            # specific models pre-defined.
            "i",
        )
    )
    VALUES = set(
        (
            "plot",
            "store",
            "resume",
            "entropy",
            "fit",
            "noise",
            "seed",
            "pars",
            "resynth",
            "time",
            "checkpoint",
            "m",
            "c",
            "p",
            "parallel",
            "view",
            "trim",
            "alpha",
            "outliers",
            # The following options are for remote fitting via the
            # fitting service, but this is not currently active.
            "transport",
            "notify",
            "queue",
        )
    )
    # Add in parameters from the fitters
    VALUES |= set(FIT_FIELDS.keys())
    # --parallel is equivalent to --parallel=0
    IMPLICIT_VALUES = {
        "parallel": "0",
        "entropy": "llf",
        "resume": "-",
    }
    pars = None
    notify = ""
    queue = None
    resynth = "0"
    noise = "5"
    starts = "1"
    seed = ""
    time = "inf"
    checkpoint = "0"
    parallel = ""
    entropy = None
    trim = "true"
    view = None
    alpha = 0.0
    PLOTTERS = "linear", "log", "residuals"
    USAGE = """\
Usage: bumps [options] modelfile [modelargs]

The modelfile is a Python script (i.e., a series of Python commands)
which sets up the data, the models, and the fittable parameters.
The model arguments are available in the modelfile as sys.argv[1:].
Model arguments may not start with '-'.

Options:

    --preview
        display model but do not perform a fitting operation
    --pars=filename or store path
        initial parameter values; fit results are saved as path/<modelname>.par
    --plot=log      [%(plotter)s]
        type of plot to display
    --trim=true
        trim any remaining burn before displaying plots [dream only]
    --simulate
        simulate a dataset using the initial problem parameters
    --simrandom
        simulate a dataset using random problem parameters
    --shake
        set random parameters before fitting
    --noise=5%%
        percent noise to add to the simulated data
    --seed=integer
        random number seed
    --err
        show uncertainty estimate from curvature at the minimum
    --cov
        show the covariance matrix for the model when done
    --entropy=gmm|mvn|wnn|llf
        compute entropy on posterior distribution [dream only]
    --staj
        output staj file when done [Refl1D only]
    --edit
        start the gui
    --view=linear|log
        one of the predefined problem views; reflectometry also has fresnel,
        logfresnel, q4 and residuals

    --store=path
        output directory for plots and models
    --overwrite
        if store already exists, replace it
    --resume=path    [dream]
        resume a fit from previous stored state; if path is '-' then use the
        path given by --store, if it exists
    --parallel=n
        run fit using multiprocessing for parallelism; use --parallel=0 for all cpus
    --mpi
        run fit using MPI for parallelism (use command "mpirun -n cpus ...")
    --batch
        batch mode; save output in .mon file and don't show plots after fit
    --noshow
        semi-batch; send output to console but don't show plots after fit
    --time=inf
        run for a maximum number of hours
    --checkpoint=0
        save fit state every n hours, or 0 for no checkpoints

    --fit=amoeba    [%(fitter)s]
        fitting engine to use; see manual for details
    --steps=0       [%(fitter)s]
        number of fit iterations after any burn-in time; use samples if steps=0
    --samples=1e4   [dream]
        set steps=samples/(pop*#pars) so the target number of samples is drawn
    --xtol=1e-4     [de, amoeba]
        minimum population diameter
    --ftol=1e-4     [de, amoeba]
        minimum population flatness
    --alpha=0.0     [dream]
        p-level for rejecting convergence; with fewer samples use a stricter
        stopping condition, such as --alpha=0.01 --samples=20000
    --outliers=none [dream]
        name of test used for removing outlier chains every N samples:
          none:   no outlier removal
          iqr:    use interquartile range on likelihood
          grubbs: use t-test on likelihood
          mahal:  use distance from parameter values on the best chain
    --pop=10        [dream, de, rl, ps]
        population size is pop times number of fitted parameters; if pop is
        negative, then set population size to -pop.
    --burn=100      [dream, pt]
        number of burn-in iterations before accumulating stats
    --thin=1        [dream]
        number of fit iterations between steps
    --nT=25
    --Tmin=0.1
    --Tmax=10       [pt]
        temperatures vector; use a higher maximum temperature and a larger
        nT if your fit is getting stuck in local minima
    --CR=0.9        [de, rl, pt]
        crossover ratio for population mixing
    --starts=1      [newton, rl, amoeba]
        number of times to run the fit from random starting points.
    --keep_best
        when running with multiple starts, restart from a point near the
        last minimum rather than using a completely random starting point.
    --init=eps      [dream]
        population initialization method:
          eps:    ball around initial parameter set
          lhs:    latin hypercube sampling
          cov:    normally distributed according to covariance matrix
          random: uniformly distributed within parameter ranges
    --stepmon
        show details for each step
    --resynth=0
        run resynthesis error analysis for n generations

    --time_model
        run the model --steps times in order to estimate total run time.
    --profile
        run the python profiler on the model; use --steps to run multiple
        models for better statistics
    --chisq
        print the model description and chisq value and exit
    -m/-c/-p command
        run the python interpreter with bumps on the path:
            m: command is a module such as bumps.cli, run as __main__
            c: command is a python one-line command
            p: command is the name of a python script
    -i
        start the interactive interpreter
    -?/-h/--help
        display this help
""" % {
        "fitter": "|".join(sorted(FIT_AVAILABLE_IDS)),
        "plotter": "|".join(PLOTTERS),
    }

    #    --remote
    #        queue fit to run on remote server
    #    --notify=user@email
    #        remote fit notification
    #    --queue=http://reflectometry.org
    #        remote job queue
    #    --transport=mp  {amqp|mp|mpi}
    #        use amqp/multiprocessing/mpi for parallel evaluation

    _plot = "log"

    def _set_plot(self, value):
        if value not in set(self.PLOTTERS):
            raise ValueError("unknown plot type %s; use %s" % (value, "|".join(self.PLOTTERS)))
        self._plot = value

    plot = property(fget=lambda self: self._plot, fset=_set_plot)
    store = None
    resume = None
    _fit = FIT_DEFAULT_ID

    @property
    def fit(self):
        return self._fit

    @fit.setter
    def fit(self, value):
        if value not in FIT_AVAILABLE_IDS:
            raise ValueError("unknown fitter %s; use %s" % (value, "|".join(sorted(FIT_AVAILABLE_IDS))))
        self._fit = value

    fit_config = FIT_CONFIG
    TRANSPORTS = "amqp", "mp", "mpi", "celery"
    _transport = "mp"

    def _set_transport(self, value):
        if value not in self.TRANSPORTS:
            raise ValueError("unknown transport %s; use %s" % (value, "|".join(self.TRANSPORTS)))
        self._transport = value

    transport = property(fget=lambda self: self._transport, fset=_set_transport)
    meshsteps = 40


def getopts():
    """
    Process command line options.

    Option values will be stored as attributes in the returned object.
    """
    opts = BumpsOpts(sys.argv)
    opts.resynth = int(opts.resynth)
    # Set a random seed if none is given; want to know the seed so we can
    # reproduce the run.  The seed needs to be saved to the monitor file.
    opts.seed = int(opts.seed) if opts.seed else np.random.randint(1000000)
    opts.fit_config.set_from_cli(opts)
    return opts



"""
MCMC model types

Usage
-----

First create a :class:`bumps.dream.bounds.Bounds` object. This stores the
ranges available on the parameters, and controls how values outside the
range are handled::

    M_bounds = bounds(minx, maxx, style='reflect|clip|fold|randomize|none')

For simple functions you can use one of the existing models.

If your model *f* computes the probability density, use :class:`Density`::

    M = Density(f, bounds=M_bounds)

If your model *f* computes the log probability density,
use :class:`LogDensity`::

    M = LogDensity(f, bounds=M_bounds)

If your model *f* computes a simulation which returns a vector, and you
have *data* associated with the simulation, use :class:`Simulation`::

    M = Simulation(f, data=data, bounds=M_bounds)

The measurement *data* can have a 1-sigma uncertainty associated with it, as
well as a *gamma* factor if the uncertainty distribution has non-Gaussian
kurtosis associated with it.


Multivariate normal distribution::

    M = MVNormal(mu, sigma)

Mixture models::

    M = Mixture(M1, w1, M2, w2, ...)


For more complex functions, you can subclass MCMCModel::

    class Model(MCMCModel):
        def __init__(self, ..., bounds=None, ...):
            ...
            self.bounds = bounds
            ...
        def nnlf(self, x):
            "Return the negative log likelihood of seeing x"
            p = probability of seeing x
            return -log(p)

    M = Model(..., bounds=M_bounds, ...)

The MCMC program uses only two methods from the model::

    apply_bounds(pop)
    log_density(pop)

If your model provides these methods, you will not need to subclass MCMCModel
in order to interact with DREAM.


Compatibility with matlab DREAM
-------------------------------

First generate a bounds handling function::

    M_bounds = bounds(ParRange.minn, ParRange.maxn)

Then generate a model, depending on what kind of function you have.

Option 1. Model directly computes posterior density::

    model = Density(f, bounds=M_bounds)

Option 2. Model computes simulation, data has known 1-sigma uncertainty::

    model = Simulation(f, data=Measurement.MeasData, bounds=M_bounds,
                       sigma=Measurement.Sigma, gamma = MCMCPar.Gamma)

Option 3. Model computes simulation, data has unknown 1-sigma uncertainty::

    model = Simulation(f, data=Measurement.MeasData, bounds=M_bounds,
                       gamma = MCMCPar.Gamma)


Option 4. Model directly computes log posterior density::

    model = LogDensity(f, bounds=M_bounds)

Option 5 is like option 2 but the reported likelihoods do not take the
1-sigma uncertainty into account.  The metropolis steps are still based
on the 1-sigma uncertainty, so use the style given in option 2 for this case.

"""

__all__ = ["MCMCModel", "Density", "LogDensity", "Simulation", "MVNormal", "Mixture"]

import numpy as np
from numpy import diag, log, exp, pi
from numpy.linalg import cholesky, inv

from . import exppow


class MCMCModel(object):
    """
    MCMCM model abstract base class.

    Each model must have a negative log likelihood function which operates
    on a point x, returning the negative log likelihood, or inf if the point
    is outside the domain.
    """

    labels = None
    bounds = None

    def nllf(self, x):
        raise NotImplemented

    def log_density(self, x):
        return -self.nllf(x)

    def plot(self, x):
        pass

    def map(self, pop):
        return np.array([self.nllf(x) for x in pop])


class Density(MCMCModel):
    """
    Construct an MCMC model from a probablility density function.

    *f* is the density function
    """

    def __init__(self, f, bounds=None, labels=None):
        self.f, self.bounds, self.labels = f, bounds, labels

    def nllf(self, x):
        return -log(self.f(x))


class LogDensity(MCMCModel):
    """
    Construct an MCMC model from a log probablility density function.

    *f* is the log density function
    """

    def __init__(self, f, bounds=None, labels=None):
        self.f, self.bounds, self.labels = f, bounds, labels

    def nllf(self, x):
        return -self.f(x)


class Simulation(MCMCModel):
    """
    Construct an MCMC model from a simulation function.

    *f* is the function which simulates the data
    *data* is the measurement(s) to compare it to
    *sigma* is the 1-sigma uncertainty of the measurement(s).
    *gamma* in (-1, 1] represents kurtosis on the data measurement uncertainty.

    Data is assumed to come from an exponential power density::

        p(v|S, G) = w(G)/S exp(-c(G) |v/S|^(2/(1+G)))

    where S is *sigma* and G is *gamma*.

    The values of *sigma* and *gamma* can be uniform or can vary with the
    individual measurement points.

    Certain values of *gamma* select particular distributions::
        G = 0: normal
        G = 1: double exponential
        G -> -1: uniform
    """

    def __init__(self, f=None, bounds=None, data=None, sigma=1, gamma=0, labels=None):
        self.f, self.bounds, self.labels = f, bounds, labels
        self.data, self.sigma, self.gamma = data, sigma, gamma
        cb, wb = exppow.exppow_pars(gamma)
        self._offset = np.sum(log(wb / sigma * np.ones_like(data)))
        self._cb = cb
        self._pow = 2 / (1 + gamma)
        # print "cb", cb, "sqrt(2pi)*wb", sqrt(2*pi)*wb
        # print "offset", self._offset

    def nllf(self, x):
        err = self.f(x) - self.data
        log_p = self._offset - np.sum(self._cb * abs(err / self.sigma) ** self._pow)
        return log_p

    def plot(self, x):
        import pylab

        v = pylab.arange(len(self.data))
        pylab.plot(v, self.data, "x", v, self.f(x), "-")


class MVNormal(MCMCModel):
    """
    multivariate normal negative log likelihood function
    """

    def __init__(self, mu, sigma):
        self.mu, self.sigma = np.asarray(mu), np.asarray(sigma)
        # Precompute sigma contributions
        r = cholesky(sigma)
        self._rinv = inv(r)
        self._c = 0.5 * len(mu) * log(2 * pi) + 0.5 * np.sum(diag(r))

    def nllf(self, x):
        mu, c, rinv = self.mu, self._c, self._rinv
        y = c + 0.5 * np.sum(np.dot(x - mu, rinv) ** 2)
        return y


class Mixture(MCMCModel):
    """
    Create a mixture model from a list of weighted density models.

    MixtureModel( M1, w1, M2, w2, ...)

    Models M1, M2, ... are MCMC models with M.nllf(x) returning the negative
    log likelihood of x.  Weights w1, w2, ... are arbitrary scalars.
    """

    def __init__(self, *args):
        models = args[::2]
        weights = args[1::2]
        if (
            len(args) % 2 != 0
            or not all(hasattr(M, "nllf") for M in models)
            or not all(np.isscalar(w) for w in weights)
        ):
            raise TypeError("Expected MixtureModel(M1, w1, M2, w2, ...)")
        self.pairs = tuple(zip(models, weights))
        self.weight = np.sum(w for w in weights)

    def nllf(self, x):
        p = [w * exp(-M.nllf(x)) for M, w in self.pairs]
        total = np.sum(p) / self.weight
        return -log(total) if total > 0 else 1e100



r"""
Freeform modeling with Chebyshev polynomials.

`Chebyshev polynomials <http://en.wikipedia.org/wiki/Chebyshev_polynomials>`_
$T_k$ form a basis set for functions over $[-1,1]$.  The truncated
interpolating polynomial $P_n$ is a weighted sum of Chebyshev polynomials
up to degree $n$:

.. math::

    f(x) \approx P_n(x) = \sum_{k=0}^n c_i T_k(x)

The interpolating polynomial exactly matches $f(x)$ at the chebyshev
nodes $z_k$ and is near the optimal polynomial approximation to $f$
of degree $n$ under the maximum norm.  For well behaved functions,
the coefficients $c_k$ decrease rapidly, and furthermore are independent
of the degree $n$ of the polynomial.

The models can either be defined directly in terms of the Chebyshev
coefficients $c_k$ with *method* = 'direct', or in terms of control
points $(z_k, f(z_k))$ at the Chebyshev nodes :func:`cheby_points`
with *method* = 'interp'.  Bounds on the parameters are easier to
control using 'interp', but the function may oscillate wildly outside
the bounds.  Bounds on the oscillation are easier to control using
'direct', but the shape of the profile is difficult to control.
"""

# TODO: clipping volume fraction to [0,1] distorts parameter space
# Option 0: clip to [0,1]
# - Bayesian analysis: parameter values outside the domain will be equally
#   probable out to infinity
# - Newton methods: the fit space is flat outside the domain, which leads
#   to a degenerate hessian.
# - Direct methods: won't fail, but will be subject to random walk
#   performance outside the domain.
# - trivial to implement!
# Option 1: compress (-inf,0.001] and [0.999,inf) into (0,0.001], [0.999,1)
# - won't address any of the problems of clipping
# Option 2: have chisq return inf for points outside the domain
# - Bayesian analysis: correctly assigns probability zero
# - Newton methods: degenerate Hessian outside domain
# - Direct methods: random walk outside domain
# - easy to implement
# Option 3: clip outside domain but add penalty based on amount of clipping
#   A profile based on clipping may have lower chisq than any profile that
#   can be described by a valid model (e.g., by having a sharper transition
#   than would be allowed by the model), leading to a minimum outside D.
#   Adding a penalty constant outside D would help, but there is no constant
#   that works everywhere.  We could use a constant greater than the worst
#   chisq seen so far in D, which can guarantee an arbitrarily low P(x) and
#   a global minimum within D, but for Newton methods, the boundary may still
#   have spurious local minima and objective value now depends on history.
#   Linear compression of profile to fit within the domain would avoid
#   unreachable profile shapes (this is just a linear transform on chebyshev
#   coefficients), and the addition of the penalty value would reduce
#   parameter correlations that result from having transformed parameters
#   resulting in identical profiles.  Returning T = ||A(x)|| from render,
#   with A being a transform that brings the profile within [0,1], the
#   objective function can return P'(x) = P(x)/(10*(1+sum(T_i)^4) for all
#   slabs i, or P(x) if no slabs return a penalty value.  So long as T is
#   monotonic with increasing badness, with value of 0 within D, and so long
#   as no values of x outside D can generate models that cannot be
#   expressed for any x within D, then any optimizer should return a valid
#   result at the global minimum.  There may still be local minima outside
#   the boundary, so information that the the value is outside the domain
#   still needs to pass through a local optimizer to the fitting program.
#   This approach could be used to transform a box constrained
#   problem to an unconstrained problem using clipping+penalty on the
#   parameter values and removing the need for constrained Newton optimizers.
# - Bayesian analysis: parameters outside D have incorrect probability, but
#   with a sufficiently large penalty, P(x) ~ 0; if the penalty value is
#   too low, details of the correlations outside D may leak into D.
# - Newton methods: Hessian should point back to domain
# - Direct methods: random walk should be biased toward the domain
# - moderately complicated
__all__ = ["profile", "cheby_approx", "cheby_val", "cheby_points", "cheby_coeff"]

import numpy as np
from numpy import real, exp, pi, cos, arange, asarray
from numpy.fft import fft


def profile(c, t, method):
    r"""
    Evaluate the chebyshev approximation c at points x.

    If method is 'direct' then $c_i$ are the coefficients for the chebyshev
    polynomials $T_i$ yielding $P = \sum_i{c_i T_i(x)}$.

    If method is 'interp' then $c_i$ are the values of the interpolated
    function $f$ evaluated at the chebyshev points returned by
    :func:`cheby_points`.
    """
    if method == "interp":
        c = cheby_coeff(c)
    return cheby_val(c, t)


def cheby_approx(n, f, range=(0, 1)):
    """
    Return the coefficients for the order n chebyshev approximation to
    function f evaluated over the range [low,high].
    """
    fx = f(cheby_points(n, range=range))
    return cheby_coeff(fx)


def cheby_val(c, x):
    r"""
    Evaluate the chebyshev approximation c at points x.

    The values $c_i$ are the coefficients for the chebyshev
    polynomials $T_i$ yielding $p(x) = \sum_i{c_i T_i(x)}$.
    """
    c = np.asarray(c)
    if len(c) == 0:
        return 0 * x

    # Crenshaw recursion from numerical recipes sec. 5.8
    y = 4 * x - 2
    d = dd = 0
    for c_j in c[:0:-1]:
        d, dd = y * d + (c_j - dd), d
    return y * (0.5 * d) + (0.5 * c[0] - dd)


def cheby_points(n, range=(0, 1)):
    r"""
    Return the points in at which a function must be evaluated to
    generate the order $n$ Chebyshev approximation function.

    Over the range [-1,1], the points are $p_k = \cos(\pi(2 k + 1)/(2n))$.
    Adjusting the range to $[x_L,x_R]$, the points become
    $x_k = \frac{1}{2} (p_k - x_L + 1)/(x_R-x_L)$.
    """
    return 0.5 * (cos(pi * (arange(n) + 0.5) / n) - range[0] + 1) / (range[1] - range[0])


def cheby_coeff(fx):
    """
    Compute chebyshev coefficients for a polynomial of order n given
    the function evaluated at the chebyshev points for order n.

    This can be used as the basis of a direct interpolation method where
    the n control points are positioned at cheby_points(n).
    """
    fx = asarray(fx)
    n = len(fx)
    w = exp((-0.5j * pi / n) * arange(n))
    y = np.hstack((fx[0::2], fx[1::2][::-1]))
    c = (2.0 / n) * real(fft(y) * w)
    return c



from .. import errplot

# from ..dream import stats as dream_stats
from ..dream import varplot as dream_varplot
from ..dream import views as dream_views
from .plot_view import PlotView


class UncertaintyView(PlotView):
    title = "Uncertainty"

    def update_parameters(self, *args, **kw):
        pass

    def plot(self):
        if not self.plot_state:
            return
        history, stats = self.plot_state
        with self.pylab_interface as pylab:
            pylab.clf()
            dream_varplot.plot_vars(history.draw(), stats)
            pylab.draw()

    def update(self, state):
        self.plot_state = state
        self.plot()

    def fit_progress(self, problem, uncertainty_state, stats):
        if problem != self.model:
            return
        self.update((uncertainty_state, stats))


class CorrelationView(PlotView):
    """
    CorrelationView has a maximum number of correlations that it will show.
    Change this by setting CorrelationView.MAX_CORR, either in the individual
    view or in the class.
    """

    MAX_CORR = 15
    title = "Correlations"

    def update_parameters(self, *args, **kw):
        pass

    def plot(self):
        if not self.plot_state:
            return
        with self.pylab_interface as pylab:
            pylab.clf()
            if self.plot_state.Nvar > self.MAX_CORR:
                return
            history = self.plot_state
            dream_views.plot_corrmatrix(history.draw())
            pylab.draw()

    def update(self, state):
        self.plot_state = state
        self.plot()

    def fit_progress(self, problem, uncertainty_state):
        if problem != self.model:
            return
        self.update(uncertainty_state)


class TraceView(PlotView):
    title = "Parameter Trace"

    def update_parameters(self, *args, **kw):
        pass

    def plot(self):
        if not self.plot_state:
            return
        history = self.plot_state
        with self.pylab_interface as pylab:
            pylab.clf()
            dream_views.plot_trace(history)
            pylab.draw()

    def update(self, state):
        self.plot_state = state
        self.plot()

    def fit_progress(self, problem, uncertainty_state):
        if problem != self.model:
            return
        self.plot_state = uncertainty_state
        self.plot()


class ModelErrorView(PlotView):
    title = "Model Uncertainty"

    def update_parameters(self, *args, **kw):
        pass

    def plot(self):
        if not self.plot_state:
            return
        with self.pylab_interface as pylab:
            pylab.clf()
            # Won't get here if plot_state is None
            errplot.show_errors(self.plot_state)
            pylab.draw()

    def update(self, problem, state):
        # TODO: Should happen in a separate process
        self.plot_state = errplot.calc_errors_from_state(problem, state)
        self.plot()

    def fit_progress(self, problem, uncertainty_state):
        if problem != self.model:
            return
        self.update(problem, uncertainty_state)



"""
Standard math functions for parameter expressions.
"""

__all__ = []

# Note: the symbols in this module are defined dynamically by parameter.py

# TODO: need a pmath test
# Something like:
#
# import numpy as np
# from bumps.parameter import Parameter, function
#
# # Define a plugin function
# @function
# def square(x): return x*x
#
# # Check all pmath symbols are imported, including plugin functions. This
# # must be done after the plugin functions are registered.
# from bumps.pmath import *
# a = Parameter(value=0.2)
# assert asind(a).value == np.degrees(np.arcsin(a.value))
# assert square(a).value = a.value**2



"""
Crossover ratios

The crossover ratio (CR) determines what percentage of parameters in the
target vector are updated with difference vector selected from the
population.  In traditional differential evolution a CR value is chosen
somewhere in [0, 1] at the start of the search and stays constant throughout.
DREAM extends this by allowing multiple CRs at the same time with different
probabilities.  Adaptive crossover adjusts the relative weights of the CRs
based on the average distance of the steps taken when that CR was used.  This
distance will be zero for unsuccessful metropolis steps, and so the relative
weights on those CRs which generate many unsuccessful steps will be reduced.

Usage
-----

1. Traditional differential evolution::

    crossover = Crossover(CR=CR)

2. Weighted crossover ratios::

    crossover = Crossover(CR=[CR1, CR2, ...], weight=[weight1, weight2, ...])

The weights are normalized to one, and default to equally weighted CRs.

3. Adaptive weighted crossover ratios::

    crossover = AdaptiveCrossover(N)

The CRs are set to *[1/N, 2/N, ... 1]*, and start out equally weighted.  The
weights are adapted during burn-in (10% of the runs) and fixed for the
remainder of the analysis.

Compatibility Notes
-------------------

For *Extra.pCR == 'Update'* in the matlab interface use::

    CR = AdaptiveCrossover(Ncr=MCMCPar.nCR)

For *Extra.pCR != 'Update'* in the matlab interface use::

    CR = Crossover(CR=[1./Ncr], pCR=[1])

"""

__all__ = [
    "Crossover",
    "BaseAdaptiveCrossover",
    "AdaptiveCrossover",
    "LogAdaptiveCrossover",
]

import numpy as np
from numpy import ones, zeros, arange, isscalar, std, trunc, log10, logspace


class Crossover(object):
    """
    Fixed weight crossover ratios.

    *CR* is a scalar if there is a single crossover ratio, or a vector of
    numbers in (0, 1].

    *weight* is the relative weighting of each CR, or None for equal weights.
    """

    def __init__(self, CR, weight=None):
        if isscalar(CR):
            CR, weight = [CR], [1]
        CR, weight = [np.asarray(v, "d") for v in (CR, weight)]
        self.CR, self.weight = CR, weight / np.sum(weight)

    def reset(self):
        pass

    def update(self, xold, xnew, used):
        """
        Gather adaptation data on *xold*, *xnew* for each CR that was
        *used* in step *N*.
        """
        pass

    def adapt(self):
        """
        Update CR weights based on the available adaptation data.
        """
        pass


class BaseAdaptiveCrossover(object):
    """
    Adapted weight crossover ratios.
    """

    weight = None  # type: np.ndarray
    _count = None  # type: np.ndarray
    _distance = None  # type: np.ndarray
    _generations = 0  # type: int
    _Nchains = 0  # type: int

    def _set_CRs(self, CR):
        self.CR = CR
        # Start with all CRs equally probable
        self.weight = ones(len(self.CR)) / len(self.CR)

        # No initial statistics for adaptation
        self._count = zeros(len(self.CR))
        self._distance = zeros(len(self.CR))
        self._generations = 0
        self._Nchains = 0

    def reset(self):
        # TODO: do we reset count and distance?
        pass

    def update(self, xold, xnew, used):
        """
        Gather adaptation data on *xold*, *xnew* for each CR that was
        *used* in step *N*.
        """
        # Calculate the standard deviation of each dimension of X
        r = std(xnew, ddof=1, axis=0)
        # Compute the Euclidean distance between new X and old X
        d = np.sum(((xold - xnew) / r) ** 2, axis=1)
        # Use this information to update sum_p2 to update N_CR
        count, total = distance_per_CR(self.CR, d, used)
        self._count += count
        self._distance += total
        self._generations += 1
        # [PAK] Accumulate number of steps so counts can be normalized
        self._Nchains += len(used)

    def adapt(self):
        """
        Update CR weights based on the available adaptation data.
        """
        # [PAK] Protect against completely stuck fits (no accepted updates
        # [PAK] and therefore a total distance of zero) by reusing the
        # [PAK] the existing weights.
        total_distance = np.sum(self._distance)
        if total_distance > 0 and self._Nchains > 0:
            # [PAK] Make sure no count is zero by adding one to all counts
            self.weight = self._distance / (self._count + 1)
            self.weight *= self._Nchains / total_distance

        # [PAK] Adjust weights toward uniform; this guarantees nonzero weights.
        self.weight += 0.1 * np.sum(self.weight)
        self.weight /= np.sum(self.weight)


class AdaptiveCrossover(BaseAdaptiveCrossover):
    """
    Adapted weight crossover ratios.

    *N* is the number of CRs to use.  CR is set to [1/N, 2/N, ..., 1], with
    initial weights [1/N, 1/N, ..., 1/N].
    """

    def __init__(self, N):
        if N < 2:
            raise ValueError("Need more than one CR for AdaptiveCrossover")
        self._set_CRs((arange(N) + 1) / N)  # Equally spaced CRs


# [PAK] Add log spaced adaptive cross-over for high dimensional tightly
# constrained problems.
class LogAdaptiveCrossover(BaseAdaptiveCrossover):
    """
    Adapted weight crossover ratios, log-spaced.

    *dim* is the number of dimensions in the problem.
    *N* is the number of CRs to use per decade.

    CR is set to [k/dim] where k is log-spaced from 1 to dim.
    The CRs start equally weighted as [1, ..., 1]/len(CR).

    *N* should be around 4.5.  This gives good low end density, with 1, 2, 3,
    and 5 parameters changed at a time, and proceeds up to 60% and 100% of
    parameters each time.  Lower values of *N* give too few high density CRs,
    and higher values give too many low density CRs.
    """

    def __init__(self, dim, N=4.5):
        # Log spaced CR from 1/dim to dim/dim
        self._set_CRs(logspace(0, log10(dim), trunc(N * log10(dim) + 1)) / dim)


def distance_per_CR(available_CRs, distances, used):
    """
    Accumulate normalized Euclidean distance for each crossover value

    Returns the number of times each available CR was used and the total
    distance for that CR.
    """
    # TODO: could use sparse array trick to evaluate totals by CR
    # Set distance[k] to coordinate (k, used[k]), then sum by columns
    # Note: currently setting unused CRs to -1, so this won't work
    total = np.asarray([np.sum(distances[used == p]) for p in available_CRs])
    count = np.asarray([np.sum(used == p) for p in available_CRs])
    return count, total
